# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import logging
from daCore import BasicObjects
import numpy, copy, itertools

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "SAMPLINGTEST")
        self.defineRequiredParameter(
            name     = "SampleAsnUplet",
            default  = [],
            typecast = tuple,
            message  = "Points de calcul définis par une liste de n-uplet",
            )
        self.defineRequiredParameter(
            name     = "SampleAsExplicitHyperCube",
            default  = [],
            typecast = tuple,
            message  = "Points de calcul définis par un hyper-cube dont on donne la liste des échantillonages de chaque variable comme une liste",
            )
        self.defineRequiredParameter(
            name     = "SampleAsMinMaxStepHyperCube",
            default  = [],
            typecast = tuple,
            message  = "Points de calcul définis par un hyper-cube dont on donne la liste des échantillonages de chaque variable par un triplet [min,max,step]",
            )
        self.defineRequiredParameter(
            name     = "SampleAsIndependantRandomVariables",
            default  = [],
            typecast = tuple,
            message  = "Points de calcul définis par un hyper-cube dont les points sur chaque axe proviennent de l'échantillonage indépendant de la variable selon la spécification ['distribution',[parametres],nombre]",
            )
        self.defineRequiredParameter(
            name     = "QualityCriterion",
            default  = "AugmentedWeightedLeastSquares",
            typecast = str,
            message  = "Critère de qualité utilisé",
            listval  = ["AugmentedWeightedLeastSquares","AWLS","AugmentedPonderatedLeastSquares","APLS","DA",
                        "WeightedLeastSquares","WLS","PonderatedLeastSquares","PLS",
                        "LeastSquares","LS","L2",
                        "AbsoluteValue","L1",
                        "MaximumError","ME"],
            )
        self.defineRequiredParameter(
            name     = "SetDebug",
            default  = False,
            typecast = bool,
            message  = "Activation du mode debug lors de l'exécution",
            )
        self.defineRequiredParameter(
            name     = "StoreSupplementaryCalculations",
            default  = [],
            typecast = tuple,
            message  = "Liste de calculs supplémentaires à stocker et/ou effectuer",
            listval  = [
                "CostFunctionJ",
                "CostFunctionJb",
                "CostFunctionJo",
                "CurrentState",
                "InnovationAtCurrentState",
                "SimulatedObservationAtCurrentState",
                ]
            )
        self.defineRequiredParameter(
            name     = "SetSeed",
            typecast = numpy.random.seed,
            message  = "Graine fixée pour le générateur aléatoire",
            )
        self.requireInputArguments(
            mandatory= ("Xb", "HO"),
            )
        self.setAttributes(tags=(
            "Checking",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, R, B, Q)
        #
        Hm = HO["Direct"].appliedTo
        #
        Xn = copy.copy( Xb )
        #
        # ---------------------------
        if len(self._parameters["SampleAsnUplet"]) > 0:
            sampleList = self._parameters["SampleAsnUplet"]
            for i,Xx in enumerate(sampleList):
                if numpy.ravel(Xx).size != Xn.size:
                    raise ValueError("The size %i of the %ith state X in the sample and %i of the checking point Xb are different, they have to be identical."%(numpy.ravel(Xx).size,i+1,Xn.size))
        elif len(self._parameters["SampleAsExplicitHyperCube"]) > 0:
            sampleList = itertools.product(*list(self._parameters["SampleAsExplicitHyperCube"]))
        elif len(self._parameters["SampleAsMinMaxStepHyperCube"]) > 0:
            coordinatesList = []
            for i,dim in enumerate(self._parameters["SampleAsMinMaxStepHyperCube"]):
                if len(dim) != 3:
                    raise ValueError("For dimension %i, the variable definition \"%s\" is incorrect, it should be [min,max,step]."%(i,dim))
                else:
                    coordinatesList.append(numpy.linspace(dim[0],dim[1],1+int((float(dim[1])-float(dim[0]))/float(dim[2]))))
            sampleList = itertools.product(*coordinatesList)
        elif len(self._parameters["SampleAsIndependantRandomVariables"]) > 0:
            coordinatesList = []
            for i,dim in enumerate(self._parameters["SampleAsIndependantRandomVariables"]):
                if len(dim) != 3:
                    raise ValueError("For dimension %i, the variable definition \"%s\" is incorrect, it should be ('distribution',(parameters),length) with distribution in ['normal'(mean,std),'lognormal'(mean,sigma),'uniform'(low,high),'weibull'(shape)]."%(i,dim))
                elif not( str(dim[0]) in ['normal','lognormal','uniform','weibull'] and hasattr(numpy.random,dim[0]) ):
                    raise ValueError("For dimension %i, the distribution name \"%s\" is not allowed, please choose in ['normal'(mean,std),'lognormal'(mean,sigma),'uniform'(low,high),'weibull'(shape)]"%(i,dim[0]))
                else:
                    distribution = getattr(numpy.random,str(dim[0]),'normal')
                    parameters   = dim[1]
                    coordinatesList.append(distribution(*dim[1], size=max(1,int(dim[2]))))
            sampleList = itertools.product(*coordinatesList)
        else:
            sampleList = iter([Xn,])
        # ----------
        BI = B.getI()
        RI = R.getI()
        def CostFunction(x, HmX, QualityMeasure="AugmentedWeightedLeastSquares"):
            if numpy.any(numpy.isnan(HmX)):
                _X  = numpy.nan
                _HX = numpy.nan
                Jb, Jo, J = numpy.nan, numpy.nan, numpy.nan
            else:
                _X  = numpy.asmatrix(numpy.ravel( x )).T
                _HX = numpy.asmatrix(numpy.ravel( HmX )).T
                if QualityMeasure in ["AugmentedWeightedLeastSquares","AWLS","AugmentedPonderatedLeastSquares","APLS","DA"]:
                    if BI is None or RI is None:
                        raise ValueError("Background and Observation error covariance matrix has to be properly defined!")
                    Jb  = 0.5 * (_X - Xb).T * BI * (_X - Xb)
                    Jo  = 0.5 * (Y - _HX).T * RI * (Y - _HX)
                elif QualityMeasure in ["WeightedLeastSquares","WLS","PonderatedLeastSquares","PLS"]:
                    if RI is None:
                        raise ValueError("Observation error covariance matrix has to be properly defined!")
                    Jb  = 0.
                    Jo  = 0.5 * (Y - _HX).T * RI * (Y - _HX)
                elif QualityMeasure in ["LeastSquares","LS","L2"]:
                    Jb  = 0.
                    Jo  = 0.5 * (Y - _HX).T * (Y - _HX)
                elif QualityMeasure in ["AbsoluteValue","L1"]:
                    Jb  = 0.
                    Jo  = numpy.sum( numpy.abs(Y - _HX) )
                elif QualityMeasure in ["MaximumError","ME"]:
                    Jb  = 0.
                    Jo  = numpy.max( numpy.abs(Y - _HX) )
                #
                J   = float( Jb ) + float( Jo )
            if self._toStore("CurrentState"):
                self.StoredVariables["CurrentState"].store( _X )
            if self._toStore("InnovationAtCurrentState"):
                self.StoredVariables["InnovationAtCurrentState"].store( Y - _HX )
            if self._toStore("SimulatedObservationAtCurrentState"):
                self.StoredVariables["SimulatedObservationAtCurrentState"].store( _HX )
            self.StoredVariables["CostFunctionJb"].store( Jb )
            self.StoredVariables["CostFunctionJo"].store( Jo )
            self.StoredVariables["CostFunctionJ" ].store( J )
            return J, Jb, Jo
        # ----------
        if self._parameters["SetDebug"]:
            CUR_LEVEL = logging.getLogger().getEffectiveLevel()
            logging.getLogger().setLevel(logging.DEBUG)
            print("===> Beginning of evaluation, activating debug\n")
            print("     %s\n"%("-"*75,))
        #
        # ----------
        for i,Xx in enumerate(sampleList):
            if self._parameters["SetDebug"]:
                print("===> Launching evaluation for state %i"%i)
            __Xn = numpy.asmatrix(numpy.ravel( Xx )).T
            try:
                Yn = Hm( __Xn )
            except:
                Yn = numpy.nan
            #
            J, Jb, Jo = CostFunction(__Xn,Yn,self._parameters["QualityCriterion"])
        # ----------
        #
        if self._parameters["SetDebug"]:
            print("\n     %s\n"%("-"*75,))
            print("===> End evaluation, deactivating debug if necessary\n")
            logging.getLogger().setLevel(CUR_LEVEL)
        #
        self._post_run(HO)
        return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

"""
    Modèles généraux pour les observers, le post-processing
"""
__author__ = "Jean-Philippe ARGAUD"
__all__ = ["ObserverTemplates"]

import numpy

# ==============================================================================
class TemplateStorage(object):
    """
    Classe générale de stockage de type dictionnaire étendu
    (Template)
    """
    def __init__( self, language = "fr_FR" ):
        self.__preferedLanguage = language
        self.__values           = {}
        self.__order            = -1

    def store( self, name = None, content = None, fr_FR = "", en_EN = "", order = "next" ):
        "D.store(k, c,  fr_FR, en_EN, o) -> Store template k and its main characteristics"
        if name is None or content is None:
            raise ValueError("To be consistent, the storage of a template must provide a name and a content.")
        if order == "next":
            self.__order += 1
        else:
            self.__order = int(order)
        self.__values[str(name)] = {
            'content': str(content),
            'fr_FR'  : str(fr_FR),
            'en_EN'  : str(en_EN),
            'order'  : int(self.__order),
            }

    def keys(self):
        "D.keys() -> list of D's keys"
        __keys = sorted(self.__values.keys())
        return __keys

    # def has_key(self, name):
    #     "D.has_key(k) -> True if D has a key k, else False"
    #     return name in self.__values

    def __contains__(self, name):
        "D.__contains__(k) -> True if D has a key k, else False"
        return name in self.__values

    def __len__(self):
        "x.__len__() <==> len(x)"
        return len(self.__values)

    def __getitem__(self, name=None ):
        "x.__getitem__(y) <==> x[y]"
        return self.__values[name]['content']

    def getdoc(self, name = None, lang = "fr_FR"):
        "D.getdoc(k, l) -> Return documentation of key k in language l"
        if lang not in self.__values[name]: lang = self.__preferedLanguage
        return self.__values[name][lang]

    def keys_in_presentation_order(self):
        "D.keys_in_presentation_order() -> list of D's keys in presentation order"
        __orders = []
        for k in self.keys():
            __orders.append( self.__values[k]['order'] )
        __reorder = numpy.array(__orders).argsort()
        return list(numpy.array(self.keys())[__reorder])

# ==============================================================================
ObserverTemplates = TemplateStorage()

ObserverTemplates.store(
    name    = "ValuePrinter",
    content = """print(str(info)+" "+str(var[-1]))""",
    fr_FR   = "Imprime sur la sortie standard la valeur courante de la variable",
    en_EN   = "Print on standard output the current value of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueAndIndexPrinter",
    content = """print(str(info)+(" index %i:"%(len(var)-1))+" "+str(var[-1]))""",
    fr_FR   = "Imprime sur la sortie standard la valeur courante de la variable, en ajoutant son index",
    en_EN   = "Print on standard output the current value of the variable, adding its index",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueSeriePrinter",
    content = """print(str(info)+" "+str(var[:]))""",
    fr_FR   = "Imprime sur la sortie standard la série des valeurs de la variable",
    en_EN   = "Print on standard output the value series of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueSaver",
    content = """import numpy, re\nv=numpy.array(var[-1], ndmin=1)\nglobal istep\ntry:\n    istep += 1\nexcept:\n    istep = 0\nf='/tmp/value_%s_%05i.txt'%(info,istep)\nf=re.sub('\\s','_',f)\nprint('Value saved in \"%s\"'%f)\nnumpy.savetxt(f,v)""",
    fr_FR   = "Enregistre la valeur courante de la variable dans un fichier du répertoire '/tmp' nommé 'value...txt' selon le nom de la variable et l'étape d'enregistrement",
    en_EN   = "Save the current value of the variable in a file of the '/tmp' directory named 'value...txt' from the variable name and the saving step",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueSerieSaver",
    content = """import numpy, re\nv=numpy.array(var[:],  ndmin=1)\nglobal istep\ntry:\n    istep += 1\nexcept:\n    istep = 0\nf='/tmp/value_%s_%05i.txt'%(info,istep)\nf=re.sub('\\s','_',f)\nprint('Value saved in \"%s\"'%f)\nnumpy.savetxt(f,v)""",
    fr_FR   = "Enregistre la série des valeurs de la variable dans un fichier du répertoire '/tmp' nommé 'value...txt' selon le nom de la variable et l'étape",
    en_EN   = "Save the value series of the variable in a file of the '/tmp' directory named 'value...txt' from the variable name and the saving step",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValuePrinterAndSaver",
    content = """import numpy, re\nv=numpy.array(var[-1], ndmin=1)\nprint(str(info)+" "+str(v))\nglobal istep\ntry:\n    istep += 1\nexcept:\n    istep = 0\nf='/tmp/value_%s_%05i.txt'%(info,istep)\nf=re.sub('\\s','_',f)\nprint('Value saved in \"%s\"'%f)\nnumpy.savetxt(f,v)""",
    fr_FR   = "Imprime sur la sortie standard et, en même temps enregistre dans un fichier du répertoire '/tmp', la valeur courante de la variable",
    en_EN   = "Print on standard output and, in the same time save in a file of the '/tmp' directory, the current value of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueIndexPrinterAndSaver",
    content = """import numpy, re\nv=numpy.array(var[-1], ndmin=1)\nprint(str(info)+(" index %i:"%(len(var)-1))+" "+str(v))\nglobal istep\ntry:\n    istep += 1\nexcept:\n    istep = 0\nf='/tmp/value_%s_%05i.txt'%(info,istep)\nf=re.sub('\\s','_',f)\nprint('Value saved in \"%s\"'%f)\nnumpy.savetxt(f,v)""",
    fr_FR   = "Imprime sur la sortie standard et, en même temps enregistre dans un fichier du répertoire '/tmp', la valeur courante de la variable, en ajoutant son index",
    en_EN   = "Print on standard output and, in the same time save in a file of the '/tmp' directory, the current value of the variable, adding its index",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueSeriePrinterAndSaver",
    content = """import numpy, re\nv=numpy.array(var[:],  ndmin=1)\nprint(str(info)+" "+str(v))\nglobal istep\ntry:\n    istep += 1\nexcept:\n    istep = 0\nf='/tmp/value_%s_%05i.txt'%(info,istep)\nf=re.sub('\\s','_',f)\nprint('Value saved in \"%s\"'%f)\nnumpy.savetxt(f,v)""",
    fr_FR   = "Imprime sur la sortie standard et, en même temps, enregistre dans un fichier du répertoire '/tmp', la série des valeurs de la variable",
    en_EN   = "Print on standard output and, in the same time, save in a file of the '/tmp' directory, the value series of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueGnuPlotter",
    content = """import numpy, Gnuplot\nv=numpy.array(var[-1], ndmin=1)\nglobal ifig, gp\ntry:\n    ifig += 1\n    gp(' set style data lines')\nexcept:\n    ifig = 0\n    gp = Gnuplot.Gnuplot(persist=1)\n    gp(' set style data lines')\ngp('set title  \"%s (Figure %i)\"'%(info,ifig))\ngp.plot( Gnuplot.Data( v, with_='lines lw 2' ) )""",
    fr_FR   = "Affiche graphiquement avec Gnuplot la valeur courante de la variable",
    en_EN   = "Graphically plot with Gnuplot the current value of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueSerieGnuPlotter",
    content = """import numpy, Gnuplot\nv=numpy.array(var[:],  ndmin=1)\nglobal ifig, gp\ntry:\n    ifig += 1\n    gp(' set style data lines')\nexcept:\n    ifig = 0\n    gp = Gnuplot.Gnuplot(persist=1)\n    gp(' set style data lines')\ngp('set title  \"%s (Figure %i)\"'%(info,ifig))\ngp.plot( Gnuplot.Data( v, with_='lines lw 2' ) )""",
    fr_FR   = "Affiche graphiquement avec Gnuplot la série des valeurs de la variable",
    en_EN   = "Graphically plot with Gnuplot the value series of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValuePrinterAndGnuPlotter",
    content = """print(str(info)+" "+str(var[-1]))\nimport numpy, Gnuplot\nv=numpy.array(var[-1], ndmin=1)\nglobal ifig,gp\ntry:\n    ifig += 1\n    gp(' set style data lines')\nexcept:\n    ifig = 0\n    gp = Gnuplot.Gnuplot(persist=1)\n    gp(' set style data lines')\ngp('set title  \"%s (Figure %i)\"'%(info,ifig))\ngp.plot( Gnuplot.Data( v, with_='lines lw 2' ) )""",
    fr_FR   = "Imprime sur la sortie standard et, en même temps, affiche graphiquement avec Gnuplot la valeur courante de la variable",
    en_EN   = "Print on standard output and, in the same time, graphically plot with Gnuplot the current value of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueSeriePrinterAndGnuPlotter",
    content = """print(str(info)+" "+str(var[:]))\nimport numpy, Gnuplot\nv=numpy.array(var[:],  ndmin=1)\nglobal ifig,gp\ntry:\n    ifig += 1\n    gp(' set style data lines')\nexcept:\n    ifig = 0\n    gp = Gnuplot.Gnuplot(persist=1)\n    gp(' set style data lines')\ngp('set title  \"%s (Figure %i)\"'%(info,ifig))\ngp.plot( Gnuplot.Data( v, with_='lines lw 2' ) )""",
    fr_FR   = "Imprime sur la sortie standard et, en même temps, affiche graphiquement avec Gnuplot la série des valeurs de la variable",
    en_EN   = "Print on standard output and, in the same time, graphically plot with Gnuplot the value series of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValuePrinterSaverAndGnuPlotter",
    content = """print(str(info)+" "+str(var[-1]))\nimport numpy, re\nv=numpy.array(var[-1], ndmin=1)\nglobal istep\ntry:\n    istep += 1\nexcept:\n    istep = 0\nf='/tmp/value_%s_%05i.txt'%(info,istep)\nf=re.sub('\\s','_',f)\nprint('Value saved in \"%s\"'%f)\nnumpy.savetxt(f,v)\nimport Gnuplot\nglobal ifig,gp\ntry:\n    ifig += 1\n    gp(' set style data lines')\nexcept:\n    ifig = 0\n    gp = Gnuplot.Gnuplot(persist=1)\n    gp(' set style data lines')\ngp('set title  \"%s (Figure %i)\"'%(info,ifig))\ngp.plot( Gnuplot.Data( v, with_='lines lw 2' ) )""",
    fr_FR   = "Imprime sur la sortie standard et, en même temps, enregistre dans un fichier du répertoire '/tmp' et affiche graphiquement la valeur courante de la variable",
    en_EN   = "Print on standard output and, in the same, time save in a file of the '/tmp' directory and graphically plot the current value of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueSeriePrinterSaverAndGnuPlotter",
    content = """print(str(info)+" "+str(var[:]))\nimport numpy, re\nv=numpy.array(var[:],  ndmin=1)\nglobal istep\ntry:\n    istep += 1\nexcept:\n    istep = 0\nf='/tmp/value_%s_%05i.txt'%(info,istep)\nf=re.sub('\\s','_',f)\nprint('Value saved in \"%s\"'%f)\nnumpy.savetxt(f,v)\nimport Gnuplot\nglobal ifig,gp\ntry:\n    ifig += 1\n    gp(' set style data lines')\nexcept:\n    ifig = 0\n    gp = Gnuplot.Gnuplot(persist=1)\n    gp(' set style data lines')\ngp('set title  \"%s (Figure %i)\"'%(info,ifig))\ngp.plot( Gnuplot.Data( v, with_='lines lw 2' ) )""",
    fr_FR   = "Imprime sur la sortie standard et, en même temps, enregistre dans un fichier du répertoire '/tmp' et affiche graphiquement la série des valeurs de la variable",
    en_EN   = "Print on standard output and, in the same, time save in a file of the '/tmp' directory and graphically plot the value series of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueMean",
    content = """import numpy\nprint(str(info)+" "+str(numpy.nanmean(var[-1])))""",
    fr_FR   = "Imprime sur la sortie standard la moyenne de la valeur courante de la variable",
    en_EN   = "Print on standard output the mean of the current value of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueStandardError",
    content = """import numpy\nprint(str(info)+" "+str(numpy.nanstd(var[-1])))""",
    fr_FR   = "Imprime sur la sortie standard l'écart-type de la valeur courante de la variable",
    en_EN   = "Print on standard output the standard error of the current value of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueVariance",
    content = """import numpy\nprint(str(info)+" "+str(numpy.nanvar(var[-1])))""",
    fr_FR   = "Imprime sur la sortie standard la variance de la valeur courante de la variable",
    en_EN   = "Print on standard output the variance of the current value of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueL2Norm",
    content = """import numpy\nv = numpy.ravel( var[-1] )\nprint(str(info)+" "+str(float( numpy.linalg.norm(v) )))""",
    fr_FR   = "Imprime sur la sortie standard la norme L2 de la valeur courante de la variable",
    en_EN   = "Print on standard output the L2 norm of the current value of the variable",
    order   = "next",
    )
ObserverTemplates.store(
    name    = "ValueRMS",
    content = """import numpy\nv = numpy.ravel( var[-1] )\nprint(str(info)+" "+str(float( numpy.sqrt((1./v.size)*numpy.dot(v,v)) )))""",
    fr_FR   = "Imprime sur la sortie standard la racine de la moyenne des carrés (RMS), ou moyenne quadratique, de la valeur courante de la variable",
    en_EN   = "Print on standard output the root mean square (RMS), or quadratic mean, of the current value of the variable",
    order   = "next",
    )

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#

__author__="aribes/gboulant"

import salome
# Get SALOME PyQt interface
import SalomePyQt
__sgPyQt = SalomePyQt.SalomePyQt()

from . import adaoModuleHelper
from PyQt5 import QtGui, QtCore
from PyQt5.QtWidgets import QApplication, QMessageBox

def waitCursor():
    QtGui.QApplication.setOverrideCursor(QtGui.QCursor(QtCore.Qt.WaitCursor))

def restoreCursor():
    QApplication.restoreOverrideCursor()

def gui_warning(parent, msg="An error occurs" ):
    """
    This function displays a message dialog box displaying the specified message.
    """
    QMessageBox.warning( parent, "Alerte", msg)

def gui_information(parent, msg="Information" ):
    """
    This function displays a message dialog box displaying the specified message.
    """
    QMessageBox.information( parent, "Information", msg, QMessageBox.Close)

#~ def getActiveStudyId():
    #~ """
    #~ This function returns the id of the active study. The concept of active study
    #~ makes sens only in the GUI context.
    #~ """
    #~ return salome.sg.getActiveStudyId()

def refreshObjectBrowser():
    """
    Refresh the graphical representation of the SALOME study, in case where the
    GUI is working.
    """
    if salome.sg is not None:
        # salome.sg.updateObjBrowser(0)
        salome.sg.updateObjBrowser()

def selectItem(salomeStudyItem):
  if salome.sg is not None:
    salome.sg.ClearIObjects()
    salome.sg.AddIObject(salomeStudyItem)

def getSelectedItem():# salomeStudyId=-100):
    """
    Get the current selection. If more than one item are selected, the
    only first is considered. The object is return (not the id).
    The item can be of any type, no control is done in this function.
    """
    if salome.sg is None:
        raise Exception("GuiHelper.getSelectedItem can't be used without the GUI context")

    #~ if salomeStudyId != -100:
      #~ studyEditor = salome.kernel.studyedit.getStudyEditor(salomeStudyId)
    studyEditor = salome.kernel.studyedit.getStudyEditor()
    item = None
    listEntries=salome.sg.getAllSelected()
    if len(listEntries) >= 1:
        entry = listEntries[0]
        item = studyEditor.study.FindObjectID( entry )
    return item

def getAllSelected():# salomeStudyId):
    """
    Returns all selected items in the specified study.
    """
    if salome.sg is None:
        raise OmaException("getSelectedItem can't be used without the GUI context", OmaException.TYPES.DEVEL)

    # study = adaoModuleHelper.getStudyManager().GetStudyByID( salomeStudyId )
    selcount = salome.sg.SelectedCount()
    seltypes = {}
    for i in range( selcount ):
        __incObjToMap( seltypes, adaoModuleHelper.getObjectID( salome.sg.getSelected( i ) ) ) # study, salome.sg.getSelected( i ) ) )
        pass
    return selcount, seltypes

def __incObjToMap( m, id ):
    """
    Increment object counter in the specified map.
    Not to be used outside this module.
    """
    if id not in m: m[id] = 0
    m[id] += 1
    pass

def getDesktop():
    """
    Returns the active Desktop. Usefull to set the relative position of a dialog box
    """
    return __sgPyQt.getDesktop()

def warning(msg):
    """
    This function displays a message dialog box displaying the specified message.
    """
    gui_warning(getDesktop(),msg)

def information(msg):
    """
    This function displays a message dialog box displaying the specified message.
    """
    gui_information(getDesktop(),msg)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

"""
    Définit des outils de persistence et d'enregistrement de séries de valeurs
    pour analyse ultérieure ou utilisation de calcul.
"""
__author__ = "Jean-Philippe ARGAUD"
__all__ = []

import os, sys, numpy, copy
import gzip, bz2

from daCore.PlatformInfo import PathManagement ; PathManagement()
from daCore.PlatformInfo import has_gnuplot, PlatformInfo
mfp = PlatformInfo().MaximumPrecision()
if has_gnuplot:
    import Gnuplot

if sys.version_info.major < 3:
    range = xrange
    iLong = long
    import cPickle as pickle
else:
    iLong = int
    import pickle

# ==============================================================================
class Persistence(object):
    """
    Classe générale de persistence définissant les accesseurs nécessaires
    (Template)
    """
    def __init__(self, name="", unit="", basetype=str):
        """
        name : nom courant
        unit : unité
        basetype : type de base de l'objet stocké à chaque pas

        La gestion interne des données est exclusivement basée sur les variables
        initialisées ici (qui ne sont pas accessibles depuis l'extérieur des
        objets comme des attributs) :
        __basetype : le type de base de chaque valeur, sous la forme d'un type
                     permettant l'instanciation ou le casting Python
        __values : les valeurs de stockage. Par défaut, c'est None
        """
        self.__name = str(name)
        self.__unit = str(unit)
        #
        self.__basetype = basetype
        #
        self.__values   = []
        self.__tags     = []
        #
        self.__dynamic  = False
        self.__g        = None
        self.__title    = None
        self.__ltitle   = None
        self.__pause    = None
        #
        self.__dataobservers = []

    def basetype(self, basetype=None):
        """
        Renvoie ou met en place le type de base des objets stockés
        """
        if basetype is None:
            return self.__basetype
        else:
            self.__basetype = basetype

    def store(self, value=None, **kwargs):
        """
        Stocke une valeur avec ses informations de filtrage.
        """
        if value is None: raise ValueError("Value argument required")
        #
        self.__values.append(copy.copy(self.__basetype(value)))
        self.__tags.append(kwargs)
        #
        if self.__dynamic: self.__replots()
        __step = len(self.__values) - 1
        for hook, parameters, scheduler in self.__dataobservers:
            if __step in scheduler:
                hook( self, parameters )

    def pop(self, item=None):
        """
        Retire une valeur enregistree par son index de stockage. Sans argument,
        retire le dernier objet enregistre.
        """
        if item is not None:
            __index = int(item)
            self.__values.pop(__index)
            self.__tags.pop(__index)
        else:
            self.__values.pop()
            self.__tags.pop()

    def shape(self):
        """
        Renvoie la taille sous forme numpy du dernier objet stocké. Si c'est un
        objet numpy, renvoie le shape. Si c'est un entier, un flottant, un
        complexe, renvoie 1. Si c'est une liste ou un dictionnaire, renvoie la
        longueur. Par défaut, renvoie 1.
        """
        if len(self.__values) > 0:
            if self.__basetype in [numpy.matrix, numpy.ndarray, numpy.array, numpy.ravel]:
                return self.__values[-1].shape
            elif self.__basetype in [int, float]:
                return (1,)
            elif self.__basetype in [list, dict]:
                return (len(self.__values[-1]),)
            else:
                return (1,)
        else:
            raise ValueError("Object has no shape before its first storage")

    # ---------------------------------------------------------
    def __str__(self):
        "x.__str__() <==> str(x)"
        msg  = "   Index        Value   Tags\n"
        for i,v in enumerate(self.__values):
            msg += "  i=%05i  %10s   %s\n"%(i,v,self.__tags[i])
        return msg

    def __len__(self):
        "x.__len__() <==> len(x)"
        return len(self.__values)

    def name(self):
        return self.__name

    def __getitem__(self, index=None ):
        "x.__getitem__(y) <==> x[y]"
        return copy.copy(self.__values[index])

    def count(self, value):
        "L.count(value) -> integer -- return number of occurrences of value"
        return self.__values.count(value)

    def index(self, value, start=0, stop=None):
        "L.index(value, [start, [stop]]) -> integer -- return first index of value."
        if stop is None : stop = len(self.__values)
        return self.__values.index(value, start, stop)

    # ---------------------------------------------------------
    def __filteredIndexes(self, **kwargs):
        "Function interne filtrant les index"
        __indexOfFilteredItems = range(len(self.__tags))
        __filteringKwTags = kwargs.keys()
        if len(__filteringKwTags) > 0:
            for tagKey in __filteringKwTags:
                __tmp = []
                for i in __indexOfFilteredItems:
                    if tagKey in self.__tags[i]:
                        if self.__tags[i][tagKey] == kwargs[tagKey]:
                            __tmp.append( i )
                        elif isinstance(kwargs[tagKey],(list,tuple)) and self.__tags[i][tagKey] in kwargs[tagKey]:
                            __tmp.append( i )
                __indexOfFilteredItems = __tmp
                if len(__indexOfFilteredItems) == 0: break
        return __indexOfFilteredItems

    # ---------------------------------------------------------
    def values(self, **kwargs):
        "D.values() -> list of D's values"
        __indexOfFilteredItems = self.__filteredIndexes(**kwargs)
        return [self.__values[i] for i in __indexOfFilteredItems]

    def keys(self, keyword=None , **kwargs):
        "D.keys() -> list of D's keys"
        __indexOfFilteredItems = self.__filteredIndexes(**kwargs)
        __keys = []
        for i in __indexOfFilteredItems:
            if keyword in self.__tags[i]:
                __keys.append( self.__tags[i][keyword] )
            else:
                __keys.append( None )
        return __keys

    def items(self, keyword=None , **kwargs):
        "D.items() -> list of D's (key, value) pairs, as 2-tuples"
        __indexOfFilteredItems = self.__filteredIndexes(**kwargs)
        __pairs = []
        for i in __indexOfFilteredItems:
            if keyword in self.__tags[i]:
                __pairs.append( (self.__tags[i][keyword], self.__values[i]) )
            else:
                __pairs.append( (None, self.__values[i]) )
        return __pairs

    def tagkeys(self):
        "D.tagkeys() -> list of D's tag keys"
        __allKeys = []
        for dicotags in self.__tags:
            __allKeys.extend( list(dicotags.keys()) )
        __allKeys = sorted(set(__allKeys))
        return __allKeys

    # def valueserie(self, item=None, allSteps=True, **kwargs):
    #     if item is not None:
    #         return self.__values[item]
    #     else:
    #         __indexOfFilteredItems = self.__filteredIndexes(**kwargs)
    #         if not allSteps and len(__indexOfFilteredItems) > 0:
    #             return self.__values[__indexOfFilteredItems[0]]
    #         else:
    #             return [self.__values[i] for i in __indexOfFilteredItems]

    def tagserie(self, item=None, withValues=False, outputTag=None, **kwargs):
        "D.tagserie() -> list of D's tag serie"
        if item is None:
            __indexOfFilteredItems = self.__filteredIndexes(**kwargs)
        else:
            __indexOfFilteredItems = [item,]
        #
        # Dans le cas où la sortie donne les valeurs d'un "outputTag"
        if outputTag is not None and isinstance(outputTag,str) :
            outputValues = []
            for index in __indexOfFilteredItems:
                if outputTag in self.__tags[index].keys():
                    outputValues.append( self.__tags[index][outputTag] )
            outputValues = sorted(set(outputValues))
            return outputValues
        #
        # Dans le cas où la sortie donne les tags satisfaisants aux conditions
        else:
            if withValues:
                return [self.__tags[index] for index in __indexOfFilteredItems]
            else:
                allTags = {}
                for index in __indexOfFilteredItems:
                    allTags.update( self.__tags[index] )
                allKeys = sorted(allTags.keys())
                return allKeys

    # ---------------------------------------------------------
    # Pour compatibilite
    def stepnumber(self):
        "Nombre de pas"
        return len(self.__values)

    # Pour compatibilite
    def stepserie(self, **kwargs):
        "Nombre de pas filtrés"
        __indexOfFilteredItems = self.__filteredIndexes(**kwargs)
        return __indexOfFilteredItems

    # Pour compatibilite
    def steplist(self, **kwargs):
        "Nombre de pas filtrés"
        __indexOfFilteredItems = self.__filteredIndexes(**kwargs)
        return list(__indexOfFilteredItems)

    # ---------------------------------------------------------
    def means(self):
        """
        Renvoie la série, contenant à chaque pas, la valeur moyenne des données
        au pas. Il faut que le type de base soit compatible avec les types
        élémentaires numpy.
        """
        try:
            return [numpy.mean(item, dtype=mfp) for item in self.__values]
        except:
            raise TypeError("Base type is incompatible with numpy")

    def stds(self, ddof=0):
        """
        Renvoie la série, contenant à chaque pas, l'écart-type des données
        au pas. Il faut que le type de base soit compatible avec les types
        élémentaires numpy.

        ddof : c'est le nombre de degrés de liberté pour le calcul de
               l'écart-type, qui est dans le diviseur. Inutile avant Numpy 1.1
        """
        try:
            if numpy.version.version >= '1.1.0':
                return [numpy.array(item).std(ddof=ddof) for item in self.__values]
            else:
                return [numpy.array(item).std() for item in self.__values]
        except:
            raise TypeError("Base type is incompatible with numpy")

    def sums(self):
        """
        Renvoie la série, contenant à chaque pas, la somme des données au pas.
        Il faut que le type de base soit compatible avec les types élémentaires
        numpy.
        """
        try:
            return [numpy.array(item).sum() for item in self.__values]
        except:
            raise TypeError("Base type is incompatible with numpy")

    def mins(self):
        """
        Renvoie la série, contenant à chaque pas, le minimum des données au pas.
        Il faut que le type de base soit compatible avec les types élémentaires
        numpy.
        """
        try:
            return [numpy.array(item).min() for item in self.__values]
        except:
            raise TypeError("Base type is incompatible with numpy")

    def maxs(self):
        """
        Renvoie la série, contenant à chaque pas, la maximum des données au pas.
        Il faut que le type de base soit compatible avec les types élémentaires
        numpy.
        """
        try:
            return [numpy.array(item).max() for item in self.__values]
        except:
            raise TypeError("Base type is incompatible with numpy")

    def __preplots(self,
                   title    = "",
                   xlabel   = "",
                   ylabel   = "",
                   ltitle   = None,
                   geometry = "600x400",
                   persist  = False,
                   pause    = True,
                  ):
        "Préparation des plots"
        #
        # Vérification de la disponibilité du module Gnuplot
        if not has_gnuplot:
            raise ImportError("The Gnuplot module is required to plot the object.")
        #
        # Vérification et compléments sur les paramètres d'entrée
        if persist:
            Gnuplot.GnuplotOpts.gnuplot_command = 'gnuplot -persist -geometry '+geometry
        else:
            Gnuplot.GnuplotOpts.gnuplot_command = 'gnuplot -geometry '+geometry
        if ltitle is None:
            ltitle = ""
        self.__g = Gnuplot.Gnuplot() # persist=1
        self.__g('set terminal '+Gnuplot.GnuplotOpts.default_term)
        self.__g('set style data lines')
        self.__g('set grid')
        self.__g('set autoscale')
        self.__g('set xlabel "'+str(xlabel)+'"')
        self.__g('set ylabel "'+str(ylabel)+'"')
        self.__title  = title
        self.__ltitle = ltitle
        self.__pause  = pause

    def plots(self,
              item     = None,
              step     = None,
              steps    = None,
              title    = "",
              xlabel   = "",
              ylabel   = "",
              ltitle   = None,
              geometry = "600x400",
              filename = "",
              dynamic  = False,
              persist  = False,
              pause    = True,
             ):
        """
        Renvoie un affichage de la valeur à chaque pas, si elle est compatible
        avec un affichage Gnuplot (donc essentiellement un vecteur). Si
        l'argument "step" existe dans la liste des pas de stockage effectués,
        renvoie l'affichage de la valeur stockée à ce pas "step". Si l'argument
        "item" est correct, renvoie l'affichage de la valeur stockée au numéro
        "item". Par défaut ou en l'absence de "step" ou "item", renvoie un
        affichage successif de tous les pas.

        Arguments :
            - step     : valeur du pas à afficher
            - item     : index de la valeur à afficher
            - steps    : liste unique des pas de l'axe des X, ou None si c'est
                         la numérotation par défaut
            - title    : base du titre général, qui sera automatiquement
                         complétée par la mention du pas
            - xlabel   : label de l'axe des X
            - ylabel   : label de l'axe des Y
            - ltitle   : titre associé au vecteur tracé
            - geometry : taille en pixels de la fenêtre et position du coin haut
                         gauche, au format X11 : LxH+X+Y (défaut : 600x400)
            - filename : base de nom de fichier Postscript pour une sauvegarde,
                         qui est automatiquement complétée par le numéro du
                         fichier calculé par incrément simple de compteur
            - dynamic  : effectue un affichage des valeurs à chaque stockage
                         (au-delà du second). La méthode "plots" permet de
                         déclarer l'affichage dynamique, et c'est la méthode
                         "__replots" qui est utilisée pour l'effectuer
            - persist  : booléen indiquant que la fenêtre affichée sera
                         conservée lors du passage au dessin suivant
                         Par défaut, persist = False
            - pause    : booléen indiquant une pause après chaque tracé, et
                         attendant un Return
                         Par défaut, pause = True
        """
        if not self.__dynamic:
            self.__preplots(title, xlabel, ylabel, ltitle, geometry, persist, pause )
            if dynamic:
                self.__dynamic = True
                if len(self.__values) == 0: return 0
        #
        # Tracé du ou des vecteurs demandés
        indexes = []
        if step is not None and step < len(self.__values):
            indexes.append(step)
        elif item is not None and item < len(self.__values):
            indexes.append(item)
        else:
            indexes = indexes + list(range(len(self.__values)))
        #
        i = -1
        for index in indexes:
            self.__g('set title  "'+str(title)+' (pas '+str(index)+')"')
            if isinstance(steps,list) or isinstance(steps,numpy.ndarray):
                Steps = list(steps)
            else:
                Steps = list(range(len(self.__values[index])))
            #
            self.__g.plot( Gnuplot.Data( Steps, self.__values[index], title=ltitle ) )
            #
            if filename != "":
                i += 1
                stepfilename = "%s_%03i.ps"%(filename,i)
                if os.path.isfile(stepfilename):
                    raise ValueError("Error: a file with this name \"%s\" already exists."%stepfilename)
                self.__g.hardcopy(filename=stepfilename, color=1)
            if self.__pause:
                eval(input('Please press return to continue...\n'))

    def __replots(self):
        """
        Affichage dans le cas du suivi dynamique de la variable
        """
        if self.__dynamic and len(self.__values) < 2: return 0
        #
        self.__g('set title  "'+str(self.__title))
        Steps = list(range(len(self.__values)))
        self.__g.plot( Gnuplot.Data( Steps, self.__values, title=self.__ltitle ) )
        #
        if self.__pause:
            eval(input('Please press return to continue...\n'))

    # ---------------------------------------------------------
    # On pourrait aussi utiliser d'autres attributs d'un "array" comme "tofile"
    def mean(self):
        """
        Renvoie la moyenne sur toutes les valeurs sans tenir compte de la
        longueur des pas. Il faut que le type de base soit compatible avec
        les types élémentaires numpy.
        """
        try:
            if self.__basetype in [int, float]:
                return float( numpy.mean(self.__values, dtype=mfp) )
            else:
                return numpy.mean(self.__values, axis=0, dtype=mfp)
        except:
            raise TypeError("Base type is incompatible with numpy")

    def std(self, ddof=0):
        """
        Renvoie l'écart-type de toutes les valeurs sans tenir compte de la
        longueur des pas. Il faut que le type de base soit compatible avec
        les types élémentaires numpy.

        ddof : c'est le nombre de degrés de liberté pour le calcul de
               l'écart-type, qui est dans le diviseur. Inutile avant Numpy 1.1
        """
        try:
            if numpy.version.version >= '1.1.0':
                return numpy.array(self.__values).std(ddof=ddof,axis=0)
            else:
                return numpy.array(self.__values).std(axis=0)
        except:
            raise TypeError("Base type is incompatible with numpy")

    def sum(self):
        """
        Renvoie la somme de toutes les valeurs sans tenir compte de la
        longueur des pas. Il faut que le type de base soit compatible avec
        les types élémentaires numpy.
        """
        try:
            return numpy.array(self.__values).sum(axis=0)
        except:
            raise TypeError("Base type is incompatible with numpy")

    def min(self):
        """
        Renvoie le minimum de toutes les valeurs sans tenir compte de la
        longueur des pas. Il faut que le type de base soit compatible avec
        les types élémentaires numpy.
        """
        try:
            return numpy.array(self.__values).min(axis=0)
        except:
            raise TypeError("Base type is incompatible with numpy")

    def max(self):
        """
        Renvoie le maximum de toutes les valeurs sans tenir compte de la
        longueur des pas. Il faut que le type de base soit compatible avec
        les types élémentaires numpy.
        """
        try:
            return numpy.array(self.__values).max(axis=0)
        except:
            raise TypeError("Base type is incompatible with numpy")

    def cumsum(self):
        """
        Renvoie la somme cumulée de toutes les valeurs sans tenir compte de la
        longueur des pas. Il faut que le type de base soit compatible avec
        les types élémentaires numpy.
        """
        try:
            return numpy.array(self.__values).cumsum(axis=0)
        except:
            raise TypeError("Base type is incompatible with numpy")

    def plot(self,
             steps    = None,
             title    = "",
             xlabel   = "",
             ylabel   = "",
             ltitle   = None,
             geometry = "600x400",
             filename = "",
             persist  = False,
             pause    = True,
            ):
        """
        Renvoie un affichage unique pour l'ensemble des valeurs à chaque pas, si
        elles sont compatibles avec un affichage Gnuplot (donc essentiellement
        un vecteur). Si l'argument "step" existe dans la liste des pas de
        stockage effectués, renvoie l'affichage de la valeur stockée à ce pas
        "step". Si l'argument "item" est correct, renvoie l'affichage de la
        valeur stockée au numéro "item".

        Arguments :
            - steps    : liste unique des pas de l'axe des X, ou None si c'est
                         la numérotation par défaut
            - title    : base du titre général, qui sera automatiquement
                         complétée par la mention du pas
            - xlabel   : label de l'axe des X
            - ylabel   : label de l'axe des Y
            - ltitle   : titre associé au vecteur tracé
            - geometry : taille en pixels de la fenêtre et position du coin haut
                         gauche, au format X11 : LxH+X+Y (défaut : 600x400)
            - filename : nom de fichier Postscript pour une sauvegarde
            - persist  : booléen indiquant que la fenêtre affichée sera
                         conservée lors du passage au dessin suivant
                         Par défaut, persist = False
            - pause    : booléen indiquant une pause après chaque tracé, et
                         attendant un Return
                         Par défaut, pause = True
        """
        #
        # Vérification de la disponibilité du module Gnuplot
        if not has_gnuplot:
            raise ImportError("The Gnuplot module is required to plot the object.")
        #
        # Vérification et compléments sur les paramètres d'entrée
        if persist:
            Gnuplot.GnuplotOpts.gnuplot_command = 'gnuplot -persist -geometry '+geometry
        else:
            Gnuplot.GnuplotOpts.gnuplot_command = 'gnuplot -geometry '+geometry
        if ltitle is None:
            ltitle = ""
        if isinstance(steps,list) or isinstance(steps, numpy.ndarray):
            Steps = list(steps)
        else:
            Steps = list(range(len(self.__values[0])))
        self.__g = Gnuplot.Gnuplot() # persist=1
        self.__g('set terminal '+Gnuplot.GnuplotOpts.default_term)
        self.__g('set style data lines')
        self.__g('set grid')
        self.__g('set autoscale')
        self.__g('set title  "'+str(title) +'"')
        self.__g('set xlabel "'+str(xlabel)+'"')
        self.__g('set ylabel "'+str(ylabel)+'"')
        #
        # Tracé du ou des vecteurs demandés
        indexes = list(range(len(self.__values)))
        self.__g.plot( Gnuplot.Data( Steps, self.__values[indexes.pop(0)], title=ltitle+" (pas 0)" ) )
        for index in indexes:
            self.__g.replot( Gnuplot.Data( Steps, self.__values[index], title=ltitle+" (pas %i)"%index ) )
        #
        if filename != "":
            self.__g.hardcopy(filename=filename, color=1)
        if pause:
            eval(input('Please press return to continue...\n'))

    # ---------------------------------------------------------
    def setDataObserver(self, HookFunction = None, HookParameters = None, Scheduler = None):
        """
        Association à la variable d'un triplet définissant un observer

        Le Scheduler attendu est une fréquence, une simple liste d'index ou un
        range des index.
        """
        #
        # Vérification du Scheduler
        # -------------------------
        maxiter = int( 1e9 )
        if isinstance(Scheduler,int):      # Considéré comme une fréquence à partir de 0
            Schedulers = range( 0, maxiter, int(Scheduler) )
        elif isinstance(Scheduler,range):  # Considéré comme un itérateur
            Schedulers = Scheduler
        elif isinstance(Scheduler,(list,tuple)):   # Considéré comme des index explicites
            Schedulers = [iLong(i) for i in Scheduler] # map( long, Scheduler )
        else:                              # Dans tous les autres cas, activé par défaut
            Schedulers = range( 0, maxiter )
        #
        # Stockage interne de l'observer dans la variable
        # -----------------------------------------------
        self.__dataobservers.append( [HookFunction, HookParameters, Schedulers] )

    def removeDataObserver(self, HookFunction = None, AllObservers = False):
        """
        Suppression d'un observer nommé sur la variable.

        On peut donner dans HookFunction la meme fonction que lors de la
        définition, ou un simple string qui est le nom de la fonction. Si
        AllObservers est vrai, supprime tous les observers enregistrés.
        """
        if hasattr(HookFunction,"func_name"):
            name = str( HookFunction.func_name )
        elif hasattr(HookFunction,"__name__"):
            name = str( HookFunction.__name__ )
        elif isinstance(HookFunction,str):
            name = str( HookFunction )
        else:
            name = None
        #
        i = -1
        index_to_remove = []
        for [hf, hp, hs] in self.__dataobservers:
            i = i + 1
            if name is hf.__name__ or AllObservers: index_to_remove.append( i )
        index_to_remove.reverse()
        for i in index_to_remove:
            self.__dataobservers.pop( i )
        return len(index_to_remove)

    def hasDataObserver(self):
        return bool(len(self.__dataobservers) > 0)

# ==============================================================================
class SchedulerTrigger(object):
    """
    Classe générale d'interface de type Scheduler/Trigger
    """
    def __init__(self,
                 simplifiedCombo = None,
                 startTime       = 0,
                 endTime         = int( 1e9 ),
                 timeDelay       = 1,
                 timeUnit        = 1,
                 frequency       = None,
                ):
        pass

# ==============================================================================
class OneScalar(Persistence):
    """
    Classe définissant le stockage d'une valeur unique réelle (float) par pas.

    Le type de base peut être changé par la méthode "basetype", mais il faut que
    le nouveau type de base soit compatible avec les types par éléments de
    numpy. On peut même utiliser cette classe pour stocker des vecteurs/listes
    ou des matrices comme dans les classes suivantes, mais c'est déconseillé
    pour conserver une signification claire des noms.
    """
    def __init__(self, name="", unit="", basetype = float):
        Persistence.__init__(self, name, unit, basetype)

class OneIndex(Persistence):
    """
    Classe définissant le stockage d'une valeur unique entière (int) par pas.
    """
    def __init__(self, name="", unit="", basetype = int):
        Persistence.__init__(self, name, unit, basetype)

class OneVector(Persistence):
    """
    Classe de stockage d'une liste de valeurs numériques homogènes par pas. Ne
    pas utiliser cette classe pour des données hétérogènes, mais "OneList".
    """
    def __init__(self, name="", unit="", basetype = numpy.ravel):
        Persistence.__init__(self, name, unit, basetype)

class OneMatrix(Persistence):
    """
    Classe de stockage d'une matrice de valeurs (numpy.matrix) par pas.
    """
    def __init__(self, name="", unit="", basetype = numpy.matrix):
        Persistence.__init__(self, name, unit, basetype)

class OneList(Persistence):
    """
    Classe de stockage d'une liste de valeurs hétérogènes (list) par pas. Ne pas
    utiliser cette classe pour des données numériques homogènes, mais
    "OneVector".
    """
    def __init__(self, name="", unit="", basetype = list):
        Persistence.__init__(self, name, unit, basetype)

def NoType( value ):
    "Fonction transparente, sans effet sur son argument"
    return value

class OneNoType(Persistence):
    """
    Classe de stockage d'un objet sans modification (cast) de type. Attention,
    selon le véritable type de l'objet stocké à chaque pas, les opérations
    arithmétiques à base de numpy peuvent être invalides ou donner des résultats
    inattendus. Cette classe n'est donc à utiliser qu'à bon escient
    volontairement, et pas du tout par défaut.
    """
    def __init__(self, name="", unit="", basetype = NoType):
        Persistence.__init__(self, name, unit, basetype)

# ==============================================================================
class CompositePersistence(object):
    """
    Structure de stockage permettant de rassembler plusieurs objets de
    persistence.

    Des objets par défaut sont prévus, et des objets supplémentaires peuvent
    être ajoutés.
    """
    def __init__(self, name="", defaults=True):
        """
        name : nom courant

        La gestion interne des données est exclusivement basée sur les variables
        initialisées ici (qui ne sont pas accessibles depuis l'extérieur des
        objets comme des attributs) :
        __StoredObjects : objets de type persistence collectés dans cet objet
        """
        self.__name = str(name)
        #
        self.__StoredObjects = {}
        #
        # Definition des objets par defaut
        # --------------------------------
        if defaults:
            self.__StoredObjects["Informations"]     = OneNoType("Informations")
            self.__StoredObjects["Background"]       = OneVector("Background", basetype=numpy.array)
            self.__StoredObjects["BackgroundError"]  = OneMatrix("BackgroundError")
            self.__StoredObjects["Observation"]      = OneVector("Observation", basetype=numpy.array)
            self.__StoredObjects["ObservationError"] = OneMatrix("ObservationError")
            self.__StoredObjects["Analysis"]         = OneVector("Analysis", basetype=numpy.array)
            self.__StoredObjects["AnalysisError"]    = OneMatrix("AnalysisError")
            self.__StoredObjects["Innovation"]       = OneVector("Innovation", basetype=numpy.array)
            self.__StoredObjects["KalmanGainK"]      = OneMatrix("KalmanGainK")
            self.__StoredObjects["OperatorH"]        = OneMatrix("OperatorH")
            self.__StoredObjects["RmsOMA"]           = OneScalar("RmsOMA")
            self.__StoredObjects["RmsOMB"]           = OneScalar("RmsOMB")
            self.__StoredObjects["RmsBMA"]           = OneScalar("RmsBMA")
        #

    def store(self, name=None, value=None, **kwargs):
        """
        Stockage d'une valeur "value" pour le "step" dans la variable "name".
        """
        if name is None: raise ValueError("Storable object name is required for storage.")
        if name not in self.__StoredObjects.keys():
            raise ValueError("No such name '%s' exists in storable objects."%name)
        self.__StoredObjects[name].store( value=value, **kwargs )

    def add_object(self, name=None, persistenceType=Persistence, basetype=None ):
        """
        Ajoute dans les objets stockables un nouvel objet défini par son nom, son
        type de Persistence et son type de base à chaque pas.
        """
        if name is None: raise ValueError("Object name is required for adding an object.")
        if name in self.__StoredObjects.keys():
            raise ValueError("An object with the same name '%s' already exists in storable objects. Choose another one."%name)
        if basetype is None:
            self.__StoredObjects[name] = persistenceType( name=str(name) )
        else:
            self.__StoredObjects[name] = persistenceType( name=str(name), basetype=basetype )

    def get_object(self, name=None ):
        """
        Renvoie l'objet de type Persistence qui porte le nom demandé.
        """
        if name is None: raise ValueError("Object name is required for retrieving an object.")
        if name not in self.__StoredObjects.keys():
            raise ValueError("No such name '%s' exists in stored objects."%name)
        return self.__StoredObjects[name]

    def set_object(self, name=None, objet=None ):
        """
        Affecte directement un 'objet' qui porte le nom 'name' demandé.
        Attention, il n'est pas effectué de vérification sur le type, qui doit
        comporter les méthodes habituelles de Persistence pour que cela
        fonctionne.
        """
        if name is None: raise ValueError("Object name is required for setting an object.")
        if name in self.__StoredObjects.keys():
            raise ValueError("An object with the same name '%s' already exists in storable objects. Choose another one."%name)
        self.__StoredObjects[name] = objet

    def del_object(self, name=None ):
        """
        Supprime un objet de la liste des objets stockables.
        """
        if name is None: raise ValueError("Object name is required for retrieving an object.")
        if name not in self.__StoredObjects.keys():
            raise ValueError("No such name '%s' exists in stored objects."%name)
        del self.__StoredObjects[name]

    # ---------------------------------------------------------
    # Méthodes d'accès de type dictionnaire
    def __getitem__(self, name=None ):
        "x.__getitem__(y) <==> x[y]"
        return self.get_object( name )

    def __setitem__(self, name=None, objet=None ):
        "x.__setitem__(i, y) <==> x[i]=y"
        self.set_object( name, objet )

    def keys(self):
        "D.keys() -> list of D's keys"
        return self.get_stored_objects(hideVoidObjects = False)

    def values(self):
        "D.values() -> list of D's values"
        return self.__StoredObjects.values()

    def items(self):
        "D.items() -> list of D's (key, value) pairs, as 2-tuples"
        return self.__StoredObjects.items()

    # ---------------------------------------------------------
    def get_stored_objects(self, hideVoidObjects = False):
        "Renvoie la liste des objets présents"
        objs = self.__StoredObjects.keys()
        if hideVoidObjects:
            usedObjs = []
            for k in objs:
                try:
                    if len(self.__StoredObjects[k]) > 0: usedObjs.append( k )
                finally:
                    pass
            objs = usedObjs
        objs = sorted(objs)
        return objs

    # ---------------------------------------------------------
    def save_composite(self, filename=None, mode="pickle", compress="gzip"):
        """
        Enregistre l'objet dans le fichier indiqué selon le "mode" demandé,
        et renvoi le nom du fichier
        """
        if filename is None:
            if compress == "gzip":
                filename = os.tempnam( os.getcwd(), 'dacp' ) + ".pkl.gz"
            elif compress == "bzip2":
                filename = os.tempnam( os.getcwd(), 'dacp' ) + ".pkl.bz2"
            else:
                filename = os.tempnam( os.getcwd(), 'dacp' ) + ".pkl"
        else:
            filename = os.path.abspath( filename )
        #
        if mode == "pickle":
            if compress == "gzip":
                output = gzip.open( filename, 'wb')
            elif compress == "bzip2":
                output = bz2.BZ2File( filename, 'wb')
            else:
                output = open( filename, 'wb')
            pickle.dump(self, output)
            output.close()
        else:
            raise ValueError("Save mode '%s' unknown. Choose another one."%mode)
        #
        return filename

    def load_composite(self, filename=None, mode="pickle", compress="gzip"):
        """
        Recharge un objet composite sauvé en fichier
        """
        if filename is None:
            raise ValueError("A file name if requested to load a composite.")
        else:
            filename = os.path.abspath( filename )
        #
        if mode == "pickle":
            if compress == "gzip":
                pkl_file = gzip.open( filename, 'rb')
            elif compress == "bzip2":
                pkl_file = bz2.BZ2File( filename, 'rb')
            else:
                pkl_file = open(filename, 'rb')
            output = pickle.load(pkl_file)
            for k in output.keys():
                self[k] = output[k]
        else:
            raise ValueError("Load mode '%s' unknown. Choose another one."%mode)
        #
        return filename

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import logging
from daCore import BasicObjects
import numpy

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "LINEARLEASTSQUARES")
        self.defineRequiredParameter(
            name     = "StoreInternalVariables",
            default  = False,
            typecast = bool,
            message  = "Stockage des variables internes ou intermédiaires du calcul",
            )
        self.defineRequiredParameter(
            name     = "StoreSupplementaryCalculations",
            default  = [],
            typecast = tuple,
            message  = "Liste de calculs supplémentaires à stocker et/ou effectuer",
            listval  = [
                "Analysis",
                "CostFunctionJ",
                "CostFunctionJAtCurrentOptimum",
                "CostFunctionJb",
                "CostFunctionJbAtCurrentOptimum",
                "CostFunctionJo",
                "CostFunctionJoAtCurrentOptimum",
                "CurrentOptimum",
                "CurrentState",
                "OMA",
                "SimulatedObservationAtCurrentOptimum",
                "SimulatedObservationAtCurrentState",
                "SimulatedObservationAtOptimum",
                ]
            )
        self.requireInputArguments(
            mandatory= ("Y", "HO", "R"),
            )
        self.setAttributes(tags=(
            "Optimization",
            "Linear",
            "Variational",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, R, B, Q)
        #
        Hm = HO["Tangent"].asMatrix(None)
        Hm = Hm.reshape(Y.size,-1) # ADAO & check shape
        Ha = HO["Adjoint"].asMatrix(None)
        Ha = Ha.reshape(-1,Y.size) # ADAO & check shape
        #
        RI = R.getI()
        #
        # Calcul de la matrice de gain et de l'analyse
        # --------------------------------------------
        K = (Ha * RI * Hm).I * Ha * RI
        Xa =  K * Y
        self.StoredVariables["Analysis"].store( Xa.A1 )
        #
        # Calcul de la fonction coût
        # --------------------------
        if self._parameters["StoreInternalVariables"] or \
            self._toStore("CostFunctionJ")  or self._toStore("CostFunctionJAtCurrentOptimum") or \
            self._toStore("CostFunctionJb") or self._toStore("CostFunctionJbAtCurrentOptimum") or \
            self._toStore("CostFunctionJo") or self._toStore("CostFunctionJoAtCurrentOptimum") or \
            self._toStore("OMA") or \
            self._toStore("SimulatedObservationAtCurrentOptimum") or \
            self._toStore("SimulatedObservationAtCurrentState") or \
            self._toStore("SimulatedObservationAtOptimum"):
            HXa = Hm * Xa
            oma = Y - HXa
        if self._parameters["StoreInternalVariables"] or \
            self._toStore("CostFunctionJ")  or self._toStore("CostFunctionJAtCurrentOptimum") or \
            self._toStore("CostFunctionJb") or self._toStore("CostFunctionJbAtCurrentOptimum") or \
            self._toStore("CostFunctionJo") or self._toStore("CostFunctionJoAtCurrentOptimum"):
            Jb  = 0.
            Jo  = float( 0.5 * oma.T * RI * oma )
            J   = Jb + Jo
            self.StoredVariables["CostFunctionJb"].store( Jb )
            self.StoredVariables["CostFunctionJo"].store( Jo )
            self.StoredVariables["CostFunctionJ" ].store( J )
            self.StoredVariables["CostFunctionJbAtCurrentOptimum"].store( Jb )
            self.StoredVariables["CostFunctionJoAtCurrentOptimum"].store( Jo )
            self.StoredVariables["CostFunctionJAtCurrentOptimum" ].store( J )
        #
        # Calculs et/ou stockages supplémentaires
        # ---------------------------------------
        if self._parameters["StoreInternalVariables"] or self._toStore("CurrentState"):
            self.StoredVariables["CurrentState"].store( numpy.ravel(Xa) )
        if self._toStore("CurrentOptimum"):
            self.StoredVariables["CurrentOptimum"].store( numpy.ravel(Xa) )
        if self._toStore("OMA"):
            self.StoredVariables["OMA"].store( numpy.ravel(oma) )
        if self._toStore("SimulatedObservationAtBackground"):
            self.StoredVariables["SimulatedObservationAtBackground"].store( numpy.ravel(HXb) )
        if self._toStore("SimulatedObservationAtCurrentState"):
            self.StoredVariables["SimulatedObservationAtCurrentState"].store( numpy.ravel(HXa) )
        if self._toStore("SimulatedObservationAtCurrentOptimum"):
            self.StoredVariables["SimulatedObservationAtCurrentOptimum"].store( numpy.ravel(HXa) )
        if self._toStore("SimulatedObservationAtOptimum"):
            self.StoredVariables["SimulatedObservationAtOptimum"].store( numpy.ravel(HXa) )
        #
        self._post_run(HO)
        return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    Définit les objets numériques génériques.
"""
__author__ = "Jean-Philippe ARGAUD"

import os, time, copy, types, sys, logging
import math, numpy, scipy
from daCore.BasicObjects import Operator
from daCore.PlatformInfo import PlatformInfo
mpr = PlatformInfo().MachinePrecision()
mfp = PlatformInfo().MaximumPrecision()
# logging.getLogger().setLevel(logging.DEBUG)

# ==============================================================================
def ExecuteFunction( paire ):
    assert len(paire) == 2, "Incorrect number of arguments"
    X, funcrepr = paire
    __X = numpy.asmatrix(numpy.ravel( X )).T
    __sys_path_tmp = sys.path ; sys.path.insert(0,funcrepr["__userFunction__path"])
    __module = __import__(funcrepr["__userFunction__modl"], globals(), locals(), [])
    __fonction = getattr(__module,funcrepr["__userFunction__name"])
    sys.path = __sys_path_tmp ; del __sys_path_tmp
    __HX  = __fonction( __X )
    return numpy.ravel( __HX )

# ==============================================================================
class FDApproximation(object):
    """
    Cette classe sert d'interface pour définir les opérateurs approximés. A la
    création d'un objet, en fournissant une fonction "Function", on obtient un
    objet qui dispose de 3 méthodes "DirectOperator", "TangentOperator" et
    "AdjointOperator". On contrôle l'approximation DF avec l'incrément
    multiplicatif "increment" valant par défaut 1%, ou avec l'incrément fixe
    "dX" qui sera multiplié par "increment" (donc en %), et on effectue de DF
    centrées si le booléen "centeredDF" est vrai.
    """
    def __init__(self,
            name                  = "FDApproximation",
            Function              = None,
            centeredDF            = False,
            increment             = 0.01,
            dX                    = None,
            avoidingRedundancy    = True,
            toleranceInRedundancy = 1.e-18,
            lenghtOfRedundancy    = -1,
            mpEnabled             = False,
            mpWorkers             = None,
            mfEnabled             = False,
            ):
        self.__name = str(name)
        if mpEnabled:
            try:
                import multiprocessing
                self.__mpEnabled = True
            except ImportError:
                self.__mpEnabled = False
        else:
            self.__mpEnabled = False
        self.__mpWorkers = mpWorkers
        if self.__mpWorkers is not None and self.__mpWorkers < 1:
            self.__mpWorkers = None
        logging.debug("FDA Calculs en multiprocessing : %s (nombre de processus : %s)"%(self.__mpEnabled,self.__mpWorkers))
        #
        if mfEnabled:
            self.__mfEnabled = True
        else:
            self.__mfEnabled = False
        logging.debug("FDA Calculs en multifonctions : %s"%(self.__mfEnabled,))
        #
        if avoidingRedundancy:
            self.__avoidRC = True
            self.__tolerBP = float(toleranceInRedundancy)
            self.__lenghtRJ = int(lenghtOfRedundancy)
            self.__listJPCP = [] # Jacobian Previous Calculated Points
            self.__listJPCI = [] # Jacobian Previous Calculated Increment
            self.__listJPCR = [] # Jacobian Previous Calculated Results
            self.__listJPPN = [] # Jacobian Previous Calculated Point Norms
            self.__listJPIN = [] # Jacobian Previous Calculated Increment Norms
        else:
            self.__avoidRC = False
        #
        if self.__mpEnabled:
            if isinstance(Function,types.FunctionType):
                logging.debug("FDA Calculs en multiprocessing : FunctionType")
                self.__userFunction__name = Function.__name__
                try:
                    mod = os.path.join(Function.__globals__['filepath'],Function.__globals__['filename'])
                except:
                    mod = os.path.abspath(Function.__globals__['__file__'])
                if not os.path.isfile(mod):
                    raise ImportError("No user defined function or method found with the name %s"%(mod,))
                self.__userFunction__modl = os.path.basename(mod).replace('.pyc','').replace('.pyo','').replace('.py','')
                self.__userFunction__path = os.path.dirname(mod)
                del mod
                self.__userOperator = Operator( name = self.__name, fromMethod = Function, avoidingRedundancy = self.__avoidRC, inputAsMultiFunction = self.__mfEnabled )
                self.__userFunction = self.__userOperator.appliedTo # Pour le calcul Direct
            elif isinstance(Function,types.MethodType):
                logging.debug("FDA Calculs en multiprocessing : MethodType")
                self.__userFunction__name = Function.__name__
                try:
                    mod = os.path.join(Function.__globals__['filepath'],Function.__globals__['filename'])
                except:
                    mod = os.path.abspath(Function.__func__.__globals__['__file__'])
                if not os.path.isfile(mod):
                    raise ImportError("No user defined function or method found with the name %s"%(mod,))
                self.__userFunction__modl = os.path.basename(mod).replace('.pyc','').replace('.pyo','').replace('.py','')
                self.__userFunction__path = os.path.dirname(mod)
                del mod
                self.__userOperator = Operator( name = self.__name, fromMethod = Function, avoidingRedundancy = self.__avoidRC, inputAsMultiFunction = self.__mfEnabled )
                self.__userFunction = self.__userOperator.appliedTo # Pour le calcul Direct
            else:
                raise TypeError("User defined function or method has to be provided for finite differences approximation.")
        else:
            self.__userOperator = Operator( name = self.__name, fromMethod = Function, avoidingRedundancy = self.__avoidRC, inputAsMultiFunction = self.__mfEnabled )
            self.__userFunction = self.__userOperator.appliedTo
        #
        self.__centeredDF = bool(centeredDF)
        if abs(float(increment)) > 1.e-15:
            self.__increment  = float(increment)
        else:
            self.__increment  = 0.01
        if dX is None:
            self.__dX     = None
        else:
            self.__dX     = numpy.asmatrix(numpy.ravel( dX )).T
        logging.debug("FDA Reduction des doublons de calcul : %s"%self.__avoidRC)
        if self.__avoidRC:
            logging.debug("FDA Tolerance de determination des doublons : %.2e"%self.__tolerBP)

    # ---------------------------------------------------------
    def __doublon__(self, e, l, n, v=None):
        __ac, __iac = False, -1
        for i in range(len(l)-1,-1,-1):
            if numpy.linalg.norm(e - l[i]) < self.__tolerBP * n[i]:
                __ac, __iac = True, i
                if v is not None: logging.debug("FDA Cas%s déja calculé, récupération du doublon %i"%(v,__iac))
                break
        return __ac, __iac

    # ---------------------------------------------------------
    def DirectOperator(self, X ):
        """
        Calcul du direct à l'aide de la fonction fournie.
        """
        logging.debug("FDA Calcul DirectOperator (explicite)")
        if self.__mfEnabled:
            _HX = self.__userFunction( X, argsAsSerie = True )
        else:
            _X = numpy.asmatrix(numpy.ravel( X )).T
            _HX = numpy.ravel(self.__userFunction( _X ))
        #
        return _HX

    # ---------------------------------------------------------
    def TangentMatrix(self, X ):
        """
        Calcul de l'opérateur tangent comme la Jacobienne par différences finies,
        c'est-à-dire le gradient de H en X. On utilise des différences finies
        directionnelles autour du point X. X est un numpy.matrix.

        Différences finies centrées (approximation d'ordre 2):
        1/ Pour chaque composante i de X, on ajoute et on enlève la perturbation
           dX[i] à la  composante X[i], pour composer X_plus_dXi et X_moins_dXi, et
           on calcule les réponses HX_plus_dXi = H( X_plus_dXi ) et HX_moins_dXi =
           H( X_moins_dXi )
        2/ On effectue les différences (HX_plus_dXi-HX_moins_dXi) et on divise par
           le pas 2*dXi
        3/ Chaque résultat, par composante, devient une colonne de la Jacobienne

        Différences finies non centrées (approximation d'ordre 1):
        1/ Pour chaque composante i de X, on ajoute la perturbation dX[i] à la
           composante X[i] pour composer X_plus_dXi, et on calcule la réponse
           HX_plus_dXi = H( X_plus_dXi )
        2/ On calcule la valeur centrale HX = H(X)
        3/ On effectue les différences (HX_plus_dXi-HX) et on divise par
           le pas dXi
        4/ Chaque résultat, par composante, devient une colonne de la Jacobienne

        """
        logging.debug("FDA Début du calcul de la Jacobienne")
        logging.debug("FDA   Incrément de............: %s*X"%float(self.__increment))
        logging.debug("FDA   Approximation centrée...: %s"%(self.__centeredDF))
        #
        if X is None or len(X)==0:
            raise ValueError("Nominal point X for approximate derivatives can not be None or void (given X: %s)."%(str(X),))
        #
        _X = numpy.asmatrix(numpy.ravel( X )).T
        #
        if self.__dX is None:
            _dX  = self.__increment * _X
        else:
            _dX = numpy.asmatrix(numpy.ravel( self.__dX )).T
        #
        if (_dX == 0.).any():
            moyenne = _dX.mean()
            if moyenne == 0.:
                _dX = numpy.where( _dX == 0., float(self.__increment), _dX )
            else:
                _dX = numpy.where( _dX == 0., moyenne, _dX )
        #
        __alreadyCalculated  = False
        if self.__avoidRC:
            __bidon, __alreadyCalculatedP = self.__doublon__(_X,  self.__listJPCP, self.__listJPPN, None)
            __bidon, __alreadyCalculatedI = self.__doublon__(_dX, self.__listJPCI, self.__listJPIN, None)
            if __alreadyCalculatedP == __alreadyCalculatedI > -1:
                __alreadyCalculated, __i = True, __alreadyCalculatedP
                logging.debug("FDA Cas J déja calculé, récupération du doublon %i"%__i)
        #
        if __alreadyCalculated:
            logging.debug("FDA   Calcul Jacobienne (par récupération du doublon %i)"%__i)
            _Jacobienne = self.__listJPCR[__i]
        else:
            logging.debug("FDA   Calcul Jacobienne (explicite)")
            if self.__centeredDF:
                #
                if self.__mpEnabled and not self.__mfEnabled:
                    funcrepr = {
                        "__userFunction__path" : self.__userFunction__path,
                        "__userFunction__modl" : self.__userFunction__modl,
                        "__userFunction__name" : self.__userFunction__name,
                    }
                    _jobs = []
                    for i in range( len(_dX) ):
                        _dXi            = _dX[i]
                        _X_plus_dXi     = numpy.array( _X.A1, dtype=float )
                        _X_plus_dXi[i]  = _X[i] + _dXi
                        _X_moins_dXi    = numpy.array( _X.A1, dtype=float )
                        _X_moins_dXi[i] = _X[i] - _dXi
                        #
                        _jobs.append( (_X_plus_dXi,  funcrepr) )
                        _jobs.append( (_X_moins_dXi, funcrepr) )
                    #
                    import multiprocessing
                    self.__pool = multiprocessing.Pool(self.__mpWorkers)
                    _HX_plusmoins_dX = self.__pool.map( ExecuteFunction, _jobs )
                    self.__pool.close()
                    self.__pool.join()
                    #
                    _Jacobienne  = []
                    for i in range( len(_dX) ):
                        _Jacobienne.append( numpy.ravel( _HX_plusmoins_dX[2*i] - _HX_plusmoins_dX[2*i+1] ) / (2.*_dX[i]) )
                    #
                elif self.__mfEnabled:
                    _xserie = []
                    for i in range( len(_dX) ):
                        _dXi            = _dX[i]
                        _X_plus_dXi     = numpy.array( _X.A1, dtype=float )
                        _X_plus_dXi[i]  = _X[i] + _dXi
                        _X_moins_dXi    = numpy.array( _X.A1, dtype=float )
                        _X_moins_dXi[i] = _X[i] - _dXi
                        #
                        _xserie.append( _X_plus_dXi )
                        _xserie.append( _X_moins_dXi )
                    #
                    _HX_plusmoins_dX = self.DirectOperator( _xserie )
                     #
                    _Jacobienne  = []
                    for i in range( len(_dX) ):
                        _Jacobienne.append( numpy.ravel( _HX_plusmoins_dX[2*i] - _HX_plusmoins_dX[2*i+1] ) / (2.*_dX[i]) )
                    #
                else:
                    _Jacobienne  = []
                    for i in range( _dX.size ):
                        _dXi            = _dX[i]
                        _X_plus_dXi     = numpy.array( _X.A1, dtype=float )
                        _X_plus_dXi[i]  = _X[i] + _dXi
                        _X_moins_dXi    = numpy.array( _X.A1, dtype=float )
                        _X_moins_dXi[i] = _X[i] - _dXi
                        #
                        _HX_plus_dXi    = self.DirectOperator( _X_plus_dXi )
                        _HX_moins_dXi   = self.DirectOperator( _X_moins_dXi )
                        #
                        _Jacobienne.append( numpy.ravel( _HX_plus_dXi - _HX_moins_dXi ) / (2.*_dXi) )
                #
            else:
                #
                if self.__mpEnabled and not self.__mfEnabled:
                    funcrepr = {
                        "__userFunction__path" : self.__userFunction__path,
                        "__userFunction__modl" : self.__userFunction__modl,
                        "__userFunction__name" : self.__userFunction__name,
                    }
                    _jobs = []
                    _jobs.append( (_X.A1, funcrepr) )
                    for i in range( len(_dX) ):
                        _X_plus_dXi    = numpy.array( _X.A1, dtype=float )
                        _X_plus_dXi[i] = _X[i] + _dX[i]
                        #
                        _jobs.append( (_X_plus_dXi, funcrepr) )
                    #
                    import multiprocessing
                    self.__pool = multiprocessing.Pool(self.__mpWorkers)
                    _HX_plus_dX = self.__pool.map( ExecuteFunction, _jobs )
                    self.__pool.close()
                    self.__pool.join()
                    #
                    _HX = _HX_plus_dX.pop(0)
                    #
                    _Jacobienne = []
                    for i in range( len(_dX) ):
                        _Jacobienne.append( numpy.ravel(( _HX_plus_dX[i] - _HX ) / _dX[i]) )
                    #
                elif self.__mfEnabled:
                    _xserie = []
                    _xserie.append( _X.A1 )
                    for i in range( len(_dX) ):
                        _X_plus_dXi    = numpy.array( _X.A1, dtype=float )
                        _X_plus_dXi[i] = _X[i] + _dX[i]
                        #
                        _xserie.append( _X_plus_dXi )
                    #
                    _HX_plus_dX = self.DirectOperator( _xserie )
                    #
                    _HX = _HX_plus_dX.pop(0)
                    #
                    _Jacobienne = []
                    for i in range( len(_dX) ):
                        _Jacobienne.append( numpy.ravel(( _HX_plus_dX[i] - _HX ) / _dX[i]) )
                   #
                else:
                    _Jacobienne  = []
                    _HX = self.DirectOperator( _X )
                    for i in range( _dX.size ):
                        _dXi            = _dX[i]
                        _X_plus_dXi     = numpy.array( _X.A1, dtype=float )
                        _X_plus_dXi[i]  = _X[i] + _dXi
                        #
                        _HX_plus_dXi = self.DirectOperator( _X_plus_dXi )
                        #
                        _Jacobienne.append( numpy.ravel(( _HX_plus_dXi - _HX ) / _dXi) )
                #
            #
            _Jacobienne = numpy.asmatrix( numpy.vstack( _Jacobienne ) ).T
            if self.__avoidRC:
                if self.__lenghtRJ < 0: self.__lenghtRJ = 2 * _X.size
                while len(self.__listJPCP) > self.__lenghtRJ:
                    self.__listJPCP.pop(0)
                    self.__listJPCI.pop(0)
                    self.__listJPCR.pop(0)
                    self.__listJPPN.pop(0)
                    self.__listJPIN.pop(0)
                self.__listJPCP.append( copy.copy(_X) )
                self.__listJPCI.append( copy.copy(_dX) )
                self.__listJPCR.append( copy.copy(_Jacobienne) )
                self.__listJPPN.append( numpy.linalg.norm(_X) )
                self.__listJPIN.append( numpy.linalg.norm(_Jacobienne) )
        #
        logging.debug("FDA Fin du calcul de la Jacobienne")
        #
        return _Jacobienne

    # ---------------------------------------------------------
    def TangentOperator(self, paire ):
        """
        Calcul du tangent à l'aide de la Jacobienne.
        """
        if self.__mfEnabled:
            assert len(paire) == 1, "Incorrect lenght of arguments"
            _paire = paire[0]
            assert len(_paire) == 2, "Incorrect number of arguments"
        else:
            assert len(paire) == 2, "Incorrect number of arguments"
            _paire = paire
        X, dX = _paire
        _Jacobienne = self.TangentMatrix( X )
        if dX is None or len(dX) == 0:
            #
            # Calcul de la forme matricielle si le second argument est None
            # -------------------------------------------------------------
            if self.__mfEnabled: return [_Jacobienne,]
            else:                return _Jacobienne
        else:
            #
            # Calcul de la valeur linéarisée de H en X appliqué à dX
            # ------------------------------------------------------
            _dX = numpy.asmatrix(numpy.ravel( dX )).T
            _HtX = numpy.dot(_Jacobienne, _dX)
            if self.__mfEnabled: return [_HtX.A1,]
            else:                return _HtX.A1

    # ---------------------------------------------------------
    def AdjointOperator(self, paire ):
        """
        Calcul de l'adjoint à l'aide de la Jacobienne.
        """
        if self.__mfEnabled:
            assert len(paire) == 1, "Incorrect lenght of arguments"
            _paire = paire[0]
            assert len(_paire) == 2, "Incorrect number of arguments"
        else:
            assert len(paire) == 2, "Incorrect number of arguments"
            _paire = paire
        X, Y = _paire
        _JacobienneT = self.TangentMatrix( X ).T
        if Y is None or len(Y) == 0:
            #
            # Calcul de la forme matricielle si le second argument est None
            # -------------------------------------------------------------
            if self.__mfEnabled: return [_JacobienneT,]
            else:                return _JacobienneT
        else:
            #
            # Calcul de la valeur de l'adjoint en X appliqué à Y
            # --------------------------------------------------
            _Y = numpy.asmatrix(numpy.ravel( Y )).T
            _HaY = numpy.dot(_JacobienneT, _Y)
            if self.__mfEnabled: return [_HaY.A1,]
            else:                return _HaY.A1

# ==============================================================================
def mmqr(
        func     = None,
        x0       = None,
        fprime   = None,
        bounds   = None,
        quantile = 0.5,
        maxfun   = 15000,
        toler    = 1.e-06,
        y        = None,
        ):
    """
    Implémentation informatique de l'algorithme MMQR, basée sur la publication :
    David R. Hunter, Kenneth Lange, "Quantile Regression via an MM Algorithm",
    Journal of Computational and Graphical Statistics, 9, 1, pp.60-77, 2000.
    """
    #
    # Recuperation des donnees et informations initiales
    # --------------------------------------------------
    variables = numpy.ravel( x0 )
    mesures   = numpy.ravel( y )
    increment = sys.float_info[0]
    p         = variables.size
    n         = mesures.size
    quantile  = float(quantile)
    #
    # Calcul des parametres du MM
    # ---------------------------
    tn      = float(toler) / n
    e0      = -tn / math.log(tn)
    epsilon = (e0-tn)/(1+math.log(e0))
    #
    # Calculs d'initialisation
    # ------------------------
    residus  = mesures - numpy.ravel( func( variables ) )
    poids    = 1./(epsilon+numpy.abs(residus))
    veps     = 1. - 2. * quantile - residus * poids
    lastsurrogate = -numpy.sum(residus*veps) - (1.-2.*quantile)*numpy.sum(residus)
    iteration = 0
    #
    # Recherche iterative
    # -------------------
    while (increment > toler) and (iteration < maxfun) :
        iteration += 1
        #
        Derivees  = numpy.array(fprime(variables))
        Derivees  = Derivees.reshape(n,p) # Necessaire pour remettre en place la matrice si elle passe par des tuyaux YACS
        DeriveesT = Derivees.transpose()
        M         =   numpy.dot( DeriveesT , (numpy.array(numpy.matrix(p*[poids,]).T)*Derivees) )
        SM        =   numpy.transpose(numpy.dot( DeriveesT , veps ))
        step      = - numpy.linalg.lstsq( M, SM, rcond=-1 )[0]
        #
        variables = variables + step
        if bounds is not None:
            while( (variables < numpy.ravel(numpy.asmatrix(bounds)[:,0])).any() or (variables > numpy.ravel(numpy.asmatrix(bounds)[:,1])).any() ):
                step      = step/2.
                variables = variables - step
        residus   = mesures - numpy.ravel( func(variables) )
        surrogate = numpy.sum(residus**2 * poids) + (4.*quantile-2.) * numpy.sum(residus)
        #
        while ( (surrogate > lastsurrogate) and ( max(list(numpy.abs(step))) > 1.e-16 ) ) :
            step      = step/2.
            variables = variables - step
            residus   = mesures - numpy.ravel( func(variables) )
            surrogate = numpy.sum(residus**2 * poids) + (4.*quantile-2.) * numpy.sum(residus)
        #
        increment     = lastsurrogate-surrogate
        poids         = 1./(epsilon+numpy.abs(residus))
        veps          = 1. - 2. * quantile - residus * poids
        lastsurrogate = -numpy.sum(residus * veps) - (1.-2.*quantile)*numpy.sum(residus)
    #
    # Mesure d'écart
    # --------------
    Ecart = quantile * numpy.sum(residus) - numpy.sum( residus[residus<0] )
    #
    return variables, Ecart, [n,p,iteration,increment,0]

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D
__doc__ = """
=====================================================
ADAO: A module for Data Assimilation and Optimization
=====================================================

About
-----

**The ADAO module provides data assimilation and optimization**
features in Python or SALOME context (see
http://www.salome-platform.org/). Briefly stated, Data Assimilation is
a methodological framework to compute the optimal estimate of the
inaccessible true value of a system state, eventually over time. It
uses information coming from experimental measurements or observations,
and from numerical *a priori* models, including information about their
errors. Parts of the framework are also known under the names of
*parameter estimation*, *inverse problems*, *Bayesian estimation*,
*optimal interpolation*, etc. More details can be found in the full
ADAO documentation (see http://www.salome-platform.org/).

Only the use of ADAO text programming interface (API/TUI) is introduced
here. This interface gives ability to create a calculation object in a
similar way than the case building obtained through the graphical
interface (GUI). When one wants to elaborate "by hand" the TUI
calculation case, it is recommended to extensively use all the ADAO
module documentation, and to go back if necessary to the graphical
interface (GUI), to get all the elements allowing to correctly set the
commands.

A simple setup example of an ADAO TUI calculation case
------------------------------------------------------

To introduce the TUI interface, lets begin by a simple but complete
example of ADAO calculation case. All the data are explicitly defined
inside the script in order to make the reading easier. The whole set of
commands is the following one::

    from numpy import array, matrix
    from adao import adaoBuilder
    case = adaoBuilder.New()
    case.set( 'AlgorithmParameters', Algorithm='3DVAR' )
    case.set( 'Background',          Vector=[0, 1, 2] )
    case.set( 'BackgroundError',     ScalarSparseMatrix=1.0 )
    case.set( 'Observation',         Vector=array([0.5, 1.5, 2.5]) )
    case.set( 'ObservationError',    DiagonalSparseMatrix='1 1 1' )
    case.set( 'ObservationOperator', Matrix='1 0 0;0 2 0;0 0 3' )
    case.set( 'Observer',            Variable="Analysis", Template="ValuePrinter" )
    case.execute()

The result of running these commands in SALOME (either as a SALOME
"*shell*" command, in the Python command window of the interface, or by
the script execution entry of the menu) is the following::

    Analysis [ 0.25000264  0.79999797  0.94999939]

More advanced examples of ADAO TUI calculation case
---------------------------------------------------

Real cases involves observations loaded from files, operators
explicitly defined as generic functions including physical simulators,
time dependant information in order to deal with forecast analysis in
addition to calibration or re-analysis. More details can be found in
the full ADAO documentation (see http://www.salome-platform.org/).

License and requirements
------------------------

The license for this module is the GNU Lesser General Public License
(Lesser GPL), as stated here and in the source files::

    <ADAO, a module for Data Assimilation and Optimization>

    Copyright (C) 2008-2020 EDF R&D

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    See http://www.salome-platform.org/

In addition, we require that all publication or presentation describing
work using this software, or all commercial or not products using it,
quote at least one of the references given below:

    * ADAO, a module for Data Assimilation and Optimization,
      http://www.salome-platform.org/

    * SALOME The Open Source Integration Platform for Numerical Simulation,
      http://www.salome-platform.org/

The documentation of the module is also covered by the license and the
requirement of quoting.
"""

import os, sys, logging
adao_py_dir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, adao_py_dir)

from daCore.version import name, version, year, date
try:
    from daYacsIntegration.daOptimizerLoop import *
except:
    logging.debug("INIT Pas de chargement initial de daOptimizerLoop")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

"""
    Normalized interface for ADAO scripting (generic API)
"""
__author__ = "Jean-Philippe ARGAUD"
__all__ = ["Aidsm"]

import os
import sys
#
from daCore.BasicObjects import State, Covariance, FullOperator, Operator
from daCore.BasicObjects import AlgorithmAndParameters, DataObserver
from daCore.BasicObjects import RegulationAndParameters, CaseLogger
from daCore import PlatformInfo
#
from daCore import ExtendedLogging ; ExtendedLogging.ExtendedLogging() # A importer en premier
import logging

# ==============================================================================
class Aidsm(object):
    """ ADAO Internal Data Structure Model """
    def __init__(self, name = "", addViewers=None):
        self.__name = str(name)
        self.__directory    = None
        self.__case = CaseLogger(self.__name, "case", addViewers)
        #
        self.__adaoObject   = {}
        self.__StoredInputs = {}
        self.__PostAnalysis = []
        #
        self.__Concepts = [
            "AlgorithmParameters",
            "Background",
            "CheckingPoint",
            "ControlInput",
            "Observation",
            "BackgroundError",
            "ObservationError",
            "EvolutionError",
            "ObservationOperator",
            "EvolutionModel",
            "ControlModel",
            "Name",
            "Directory",
            "Debug",
            "NoDebug",
            "RegulationParameters",
            "Observer",
            ]
        #
        for ename in self.__Concepts:
            self.__adaoObject[ename] = None
        for ename in ("ObservationOperator", "EvolutionModel", "ControlModel"):
            self.__adaoObject[ename] = {}
        for ename in ("Observer",):
            self.__adaoObject[ename]   = []
            self.__StoredInputs[ename] = []
        self.__StoredInputs["Name"] = self.__name
        self.__StoredInputs["Directory"] = self.__directory
        #
        # Récupère le chemin du répertoire parent et l'ajoute au path
        # (Cela complète l'action de la classe PathManagement dans PlatformInfo,
        # qui est activée dans Persistence)
        self.__parent = os.path.abspath(os.path.join(os.path.dirname(__file__),".."))
        sys.path.insert(0, self.__parent)
        sys.path = PlatformInfo.uniq( sys.path ) # Conserve en unique exemplaire chaque chemin

    def set(self,
            Concept              = None, # Premier argument
            Algorithm            = None,
            AppliedInXb          = None,
            AvoidRC              = True,
            Checked              = False,
            ColMajor             = False,
            ColNames             = None,
            DataFile             = None,
            DiagonalSparseMatrix = None,
            ExtraArguments       = None,
            Info                 = None,
            InputFunctionAsMulti = False,
            Matrix               = None,
            ObjectFunction       = None,
            ObjectMatrix         = None,
            OneFunction          = None,
            Parameters           = None,
            ScalarSparseMatrix   = None,
            Scheduler            = None,
            Script               = None,
            Stored               = False,
            String               = None,
            Template             = None,
            ThreeFunctions       = None,
            Variable             = None,
            Vector               = None,
            VectorSerie          = None,
            ):
        "Interface unique de definition de variables d'entrees par argument"
        self.__case.register("set",dir(),locals(),None,True)
        try:
            if   Concept in ("Background", "CheckingPoint", "ControlInput", "Observation"):
                commande = getattr(self,"set"+Concept)
                commande(Vector, VectorSerie, Script, DataFile, ColNames, ColMajor, Stored, Scheduler, Checked )
            elif Concept in ("BackgroundError", "ObservationError", "EvolutionError"):
                commande = getattr(self,"set"+Concept)
                commande(Matrix, ScalarSparseMatrix, DiagonalSparseMatrix,
                         Script, Stored, ObjectMatrix, Checked )
            elif Concept == "AlgorithmParameters":
                self.setAlgorithmParameters( Algorithm, Parameters, Script )
            elif Concept == "RegulationParameters":
                self.setRegulationParameters( Algorithm, Parameters, Script )
            elif Concept == "Name":
                self.setName(String)
            elif Concept == "Directory":
                self.setDirectory(String)
            elif Concept == "Debug":
                self.setDebug()
            elif Concept == "NoDebug":
                self.setNoDebug()
            elif Concept == "Observer":
                self.setObserver( Variable, Template, String, Script, Info, ObjectFunction, Scheduler )
            elif Concept == "ObservationOperator":
                self.setObservationOperator(
                    Matrix, OneFunction, ThreeFunctions, AppliedInXb,
                    Parameters, Script, ExtraArguments,
                    Stored, AvoidRC, InputFunctionAsMulti, Checked )
            elif Concept in ("EvolutionModel", "ControlModel"):
                commande = getattr(self,"set"+Concept)
                commande(
                    Matrix, OneFunction, ThreeFunctions,
                    Parameters, Script, Scheduler, ExtraArguments,
                    Stored, AvoidRC, InputFunctionAsMulti, Checked )
            else:
                raise ValueError("the variable named '%s' is not allowed."%str(Concept))
        except Exception as e:
            if isinstance(e, SyntaxError): msg = "at %s: %s"%(e.offset, e.text)
            else: msg = ""
            raise ValueError(("during settings, the following error occurs:\n"+\
                              "\n%s %s\n\nSee also the potential messages, "+\
                              "which can show the origin of the above error, "+\
                              "in the launching terminal.")%(str(e),msg))

    # -----------------------------------------------------------

    def setBackground(self,
            Vector         = None,
            VectorSerie    = None,
            Script         = None,
            DataFile       = None,
            ColNames       = None,
            ColMajor       = False,
            Stored         = False,
            Scheduler      = None,
            Checked        = False):
        "Definition d'un concept de calcul"
        Concept = "Background"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = State(
            name               = Concept,
            asVector           = Vector,
            asPersistentVector = VectorSerie,
            asScript           = self.__with_directory(Script),
            asDataFile         = DataFile,
            colNames           = ColNames,
            colMajor           = ColMajor,
            scheduledBy        = Scheduler,
            toBeChecked        = Checked,
            )
        if Stored:
            self.__StoredInputs[Concept] = self.__adaoObject[Concept].getO()
        return 0

    def setCheckingPoint(self,
            Vector         = None,
            VectorSerie    = None,
            Script         = None,
            DataFile       = None,
            ColNames       = None,
            ColMajor       = False,
            Stored         = False,
            Scheduler      = None,
            Checked        = False):
        "Definition d'un concept de calcul"
        Concept = "CheckingPoint"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = State(
            name               = Concept,
            asVector           = Vector,
            asPersistentVector = VectorSerie,
            asScript           = self.__with_directory(Script),
            asDataFile         = DataFile,
            colNames           = ColNames,
            colMajor           = ColMajor,
            scheduledBy        = Scheduler,
            toBeChecked        = Checked,
            )
        if Stored:
            self.__StoredInputs[Concept] = self.__adaoObject[Concept].getO()
        return 0

    def setControlInput(self,
            Vector         = None,
            VectorSerie    = None,
            Script         = None,
            DataFile       = None,
            ColNames       = None,
            ColMajor       = False,
            Stored         = False,
            Scheduler      = None,
            Checked        = False):
        "Definition d'un concept de calcul"
        Concept = "ControlInput"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = State(
            name               = Concept,
            asVector           = Vector,
            asPersistentVector = VectorSerie,
            asScript           = self.__with_directory(Script),
            asDataFile         = DataFile,
            colNames           = ColNames,
            colMajor           = ColMajor,
            scheduledBy        = Scheduler,
            toBeChecked        = Checked,
            )
        if Stored:
            self.__StoredInputs[Concept] = self.__adaoObject[Concept].getO()
        return 0

    def setObservation(self,
            Vector         = None,
            VectorSerie    = None,
            Script         = None,
            DataFile       = None,
            ColNames       = None,
            ColMajor       = False,
            Stored         = False,
            Scheduler      = None,
            Checked        = False):
        "Definition d'un concept de calcul"
        Concept = "Observation"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = State(
            name               = Concept,
            asVector           = Vector,
            asPersistentVector = VectorSerie,
            asScript           = self.__with_directory(Script),
            asDataFile         = DataFile,
            colNames           = ColNames,
            colMajor           = ColMajor,
            scheduledBy        = Scheduler,
            toBeChecked        = Checked,
            )
        if Stored:
            self.__StoredInputs[Concept] = self.__adaoObject[Concept].getO()
        return 0

    def setBackgroundError(self,
            Matrix               = None,
            ScalarSparseMatrix   = None,
            DiagonalSparseMatrix = None,
            Script               = None,
            Stored               = False,
            ObjectMatrix         = None,
            Checked              = False):
        "Definition d'un concept de calcul"
        Concept = "BackgroundError"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = Covariance(
            name          = Concept,
            asCovariance  = Matrix,
            asEyeByScalar = ScalarSparseMatrix,
            asEyeByVector = DiagonalSparseMatrix,
            asCovObject   = ObjectMatrix,
            asScript      = self.__with_directory(Script),
            toBeChecked   = Checked,
            )
        if Stored:
            self.__StoredInputs[Concept] = self.__adaoObject[Concept].getO()
        return 0

    def setObservationError(self,
            Matrix               = None,
            ScalarSparseMatrix   = None,
            DiagonalSparseMatrix = None,
            Script               = None,
            Stored               = False,
            ObjectMatrix         = None,
            Checked              = False):
        "Definition d'un concept de calcul"
        Concept = "ObservationError"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = Covariance(
            name          = Concept,
            asCovariance  = Matrix,
            asEyeByScalar = ScalarSparseMatrix,
            asEyeByVector = DiagonalSparseMatrix,
            asCovObject   = ObjectMatrix,
            asScript      = self.__with_directory(Script),
            toBeChecked   = Checked,
            )
        if Stored:
            self.__StoredInputs[Concept] = self.__adaoObject[Concept].getO()
        return 0

    def setEvolutionError(self,
            Matrix               = None,
            ScalarSparseMatrix   = None,
            DiagonalSparseMatrix = None,
            Script               = None,
            Stored               = False,
            ObjectMatrix         = None,
            Checked              = False):
        "Definition d'un concept de calcul"
        Concept = "EvolutionError"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = Covariance(
            name          = Concept,
            asCovariance  = Matrix,
            asEyeByScalar = ScalarSparseMatrix,
            asEyeByVector = DiagonalSparseMatrix,
            asCovObject   = ObjectMatrix,
            asScript      = self.__with_directory(Script),
            toBeChecked   = Checked,
            )
        if Stored:
            self.__StoredInputs[Concept] = self.__adaoObject[Concept].getO()
        return 0

    def setObservationOperator(self,
            Matrix               = None,
            OneFunction          = None,
            ThreeFunctions       = None,
            AppliedInXb          = None,
            Parameters           = None,
            Script               = None,
            ExtraArguments       = None,
            Stored               = False,
            AvoidRC              = True,
            InputFunctionAsMulti = False,
            Checked              = False):
        "Definition d'un concept de calcul"
        Concept = "ObservationOperator"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = FullOperator(
            name             = Concept,
            asMatrix         = Matrix,
            asOneFunction    = OneFunction,
            asThreeFunctions = ThreeFunctions,
            asScript         = self.__with_directory(Script),
            asDict           = Parameters,
            appliedInX       = AppliedInXb,
            extraArguments   = ExtraArguments,
            avoidRC          = AvoidRC,
            inputAsMF        = InputFunctionAsMulti,
            scheduledBy      = None,
            toBeChecked      = Checked,
            )
        if Stored:
            self.__StoredInputs[Concept] = self.__adaoObject[Concept].getO()
        return 0

    def setEvolutionModel(self,
            Matrix               = None,
            OneFunction          = None,
            ThreeFunctions       = None,
            Parameters           = None,
            Script               = None,
            Scheduler            = None,
            ExtraArguments       = None,
            Stored               = False,
            AvoidRC              = True,
            InputFunctionAsMulti = False,
            Checked              = False):
        "Definition d'un concept de calcul"
        Concept = "EvolutionModel"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = FullOperator(
            name             = Concept,
            asMatrix         = Matrix,
            asOneFunction    = OneFunction,
            asThreeFunctions = ThreeFunctions,
            asScript         = self.__with_directory(Script),
            asDict           = Parameters,
            appliedInX       = None,
            extraArguments   = ExtraArguments,
            avoidRC          = AvoidRC,
            inputAsMF        = InputFunctionAsMulti,
            scheduledBy      = Scheduler,
            toBeChecked      = Checked,
            )
        if Stored:
            self.__StoredInputs[Concept] = self.__adaoObject[Concept].getO()
        return 0

    def setControlModel(self,
            Matrix               = None,
            OneFunction          = None,
            ThreeFunctions       = None,
            Parameters           = None,
            Script               = None,
            Scheduler            = None,
            ExtraArguments       = None,
            Stored               = False,
            AvoidRC              = True,
            InputFunctionAsMulti = False,
            Checked              = False):
        "Definition d'un concept de calcul"
        Concept = "ControlModel"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = FullOperator(
            name             = Concept,
            asMatrix         = Matrix,
            asOneFunction    = OneFunction,
            asThreeFunctions = ThreeFunctions,
            asScript         = self.__with_directory(Script),
            asDict           = Parameters,
            appliedInX       = None,
            extraArguments   = ExtraArguments,
            avoidRC          = AvoidRC,
            inputAsMF        = InputFunctionAsMulti,
            scheduledBy      = Scheduler,
            toBeChecked      = Checked,
            )
        if Stored:
            self.__StoredInputs[Concept] = self.__adaoObject[Concept].getO()
        return 0

    def setName(self, String=None):
        "Definition d'un concept de calcul"
        self.__case.register("setName",dir(),locals())
        if String is not None:
            self.__name = str(String)
        else:
            self.__name = None
        self.__StoredInputs["Name"] = self.__name

    def setDirectory(self, String=None):
        "Definition d'un concept de calcul"
        self.__case.register("setDirectory",dir(),locals())
        if os.path.isdir(os.path.abspath(str(String))):
            self.__directory = os.path.abspath(str(String))
        else:
            self.__directory = None
        self.__StoredInputs["Directory"] = self.__directory

    def setDebug(self, __level = 10):
        "NOTSET=0 < DEBUG=10 < INFO=20 < WARNING=30 < ERROR=40 < CRITICAL=50"
        self.__case.register("setDebug",dir(),locals())
        log = logging.getLogger()
        log.setLevel( __level )
        self.__StoredInputs["Debug"]   = __level
        self.__StoredInputs["NoDebug"] = False
        return 0

    def setNoDebug(self):
        "NOTSET=0 < DEBUG=10 < INFO=20 < WARNING=30 < ERROR=40 < CRITICAL=50"
        self.__case.register("setNoDebug",dir(),locals())
        log = logging.getLogger()
        log.setLevel( logging.WARNING )
        self.__StoredInputs["Debug"]   = logging.WARNING
        self.__StoredInputs["NoDebug"] = True
        return 0

    def setAlgorithmParameters(self,
            Algorithm  = None,
            Parameters = None,
            Script     = None):
        "Definition d'un concept de calcul"
        Concept = "AlgorithmParameters"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = AlgorithmAndParameters(
            name          = Concept,
            asAlgorithm   = Algorithm,
            asDict        = Parameters,
            asScript      = self.__with_directory(Script),
            )
        return 0

    def updateAlgorithmParameters(self,
            Parameters = None,
            Script     = None):
        "Mise a jour d'un concept de calcul"
        if "AlgorithmParameters" not in self.__adaoObject or self.__adaoObject["AlgorithmParameters"] is None:
            raise ValueError("\n\nNo algorithm registred, set one before updating parameters or executing\n")
        self.__adaoObject["AlgorithmParameters"].updateParameters(
            asDict        = Parameters,
            asScript      = self.__with_directory(Script),
            )
        return 0

    def setRegulationParameters(self,
            Algorithm  = None,
            Parameters = None,
            Script     = None):
        "Definition d'un concept de calcul"
        Concept = "RegulationParameters"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept] = RegulationAndParameters(
            name          = Concept,
            asAlgorithm   = Algorithm,
            asDict        = Parameters,
            asScript      = self.__with_directory(Script),
            )
        return 0

    def setObserver(self,
            Variable       = None,
            Template       = None,
            String         = None,
            Script         = None,
            Info           = None,
            ObjectFunction = None,
            Scheduler      = None):
        "Definition d'un concept de calcul"
        Concept = "Observer"
        self.__case.register("set"+Concept, dir(), locals())
        self.__adaoObject[Concept].append( DataObserver(
            name        = Concept,
            onVariable  = Variable,
            asTemplate  = Template,
            asString    = String,
            asScript    = self.__with_directory(Script),
            asObsObject = ObjectFunction,
            withInfo    = Info,
            scheduledBy = Scheduler,
            withAlgo    = self.__adaoObject["AlgorithmParameters"]
            ))
        return 0

    def removeObserver(self,
            Variable       = None,
            ObjectFunction = None,
            ):
        "Permet de retirer un observer à une ou des variable nommée"
        if "AlgorithmParameters" not in self.__adaoObject:
            raise ValueError("No algorithm registred, ask for one before removing observers")
        #
        # Vérification du nom de variable et typage
        # -----------------------------------------
        if isinstance(Variable, str):
            VariableNames = (Variable,)
        elif isinstance(Variable, list):
            VariableNames = tuple(map( str, Variable ))
        else:
            raise ValueError("The observer requires a name or a list of names of variables.")
        #
        # Association interne de l'observer à la variable
        # -----------------------------------------------
        for ename in VariableNames:
            if ename not in self.__adaoObject["AlgorithmParameters"]:
                raise ValueError("An observer requires to be removed on a variable named %s which does not exist."%ename)
            else:
                return self.__adaoObject["AlgorithmParameters"].removeObserver( ename, ObjectFunction )

    # -----------------------------------------------------------

    def get(self, Concept=None, noDetails=True ):
        "Recuperation d'une sortie du calcul"
        if Concept is not None:
            try:
                self.__case.register("get", dir(), locals(), Concept) # Break pickle in Python 2
            except Exception:
                pass
            if Concept in self.__StoredInputs:
                return self.__StoredInputs[Concept]
                #
            elif self.__adaoObject["AlgorithmParameters"] is not None and Concept == "AlgorithmParameters":
                return self.__adaoObject["AlgorithmParameters"].get()
                #
            elif self.__adaoObject["AlgorithmParameters"] is not None and Concept in self.__adaoObject["AlgorithmParameters"]:
                return self.__adaoObject["AlgorithmParameters"].get( Concept )
                #
            elif Concept == "AlgorithmRequiredParameters" and self.__adaoObject["AlgorithmParameters"] is not None:
                return self.__adaoObject["AlgorithmParameters"].getAlgorithmRequiredParameters(noDetails)
                #
            elif Concept == "AlgorithmRequiredInputs" and self.__adaoObject["AlgorithmParameters"] is not None:
                return self.__adaoObject["AlgorithmParameters"].getAlgorithmInputArguments()
                #
            elif Concept == "AlgorithmAttributes" and self.__adaoObject["AlgorithmParameters"] is not None:
                return self.__adaoObject["AlgorithmParameters"].getAlgorithmAttributes()
                #
            else:
                raise ValueError("The requested key \"%s\" does not exists as an input or a stored variable."%Concept)
        else:
            allvariables = {}
            allvariables.update( {"AlgorithmParameters":self.__adaoObject["AlgorithmParameters"].get()} )
            # allvariables.update( self.__adaoObject["AlgorithmParameters"].get() )
            allvariables.update( self.__StoredInputs )
            allvariables.pop('Observer', None)
            return allvariables

    # -----------------------------------------------------------

    def get_available_variables(self):
        """
        Renvoie les variables potentiellement utilisables pour l'étude,
        initialement stockées comme données d'entrées ou dans les algorithmes,
        identifiés par les chaînes de caractères. L'algorithme doit avoir été
        préalablement choisi sinon la méthode renvoie "None".
        """
        if len(list(self.__adaoObject["AlgorithmParameters"].keys())) == 0 and \
            len(list(self.__StoredInputs.keys())) == 0:
            return None
        else:
            variables = []
            if len(list(self.__adaoObject["AlgorithmParameters"].keys())) > 0:
                variables.extend(list(self.__adaoObject["AlgorithmParameters"].keys()))
            if len(list(self.__StoredInputs.keys())) > 0:
                variables.extend( list(self.__StoredInputs.keys()) )
            variables.remove('Observer')
            variables.sort()
            return variables

    def get_available_algorithms(self):
        """
        Renvoie la liste des algorithmes potentiellement utilisables, identifiés
        par les chaînes de caractères.
        """
        files = []
        for directory in sys.path:
            trypath = os.path.join(directory,"daAlgorithms")
            if os.path.isdir(trypath):
                for fname in os.listdir(trypath):
                    if os.path.isfile(os.path.join(trypath,fname)):
                        root, ext = os.path.splitext(fname)
                        if ext != ".py": continue
                        fc = open(os.path.join(trypath,fname)).read()
                        iselal = bool("class ElementaryAlgorithm" in fc)
                        if iselal and ext == '.py' and root != '__init__':
                            files.append(root)
        files.sort()
        return files

    def get_algorithms_main_path(self):
        """
        Renvoie le chemin pour le répertoire principal contenant les algorithmes
        """
        return self.__parent

    def add_algorithms_path(self, Path=None):
        """
        Ajoute au chemin de recherche des algorithmes un répertoire dans lequel
        se trouve un sous-répertoire "daAlgorithms"
        """
        if not os.path.isdir(Path):
            raise ValueError("The given "+Path+" argument must exist as a directory")
        if not os.path.isdir(os.path.join(Path,"daAlgorithms")):
            raise ValueError("The given \""+Path+"\" argument must contain a subdirectory named \"daAlgorithms\"")
        if not os.path.isfile(os.path.join(Path,"daAlgorithms","__init__.py")):
            raise ValueError("The given \""+Path+"/daAlgorithms\" path must contain a file named \"__init__.py\"")
        sys.path.insert(0, os.path.abspath(Path))
        sys.path = PlatformInfo.uniq( sys.path ) # Conserve en unique exemplaire chaque chemin
        return 0

    # -----------------------------------------------------------

    def execute(self, Executor=None, SaveCaseInFile=None, nextStep=False):
        "Lancement du calcul"
        self.__case.register("execute",dir(),locals(),None,True)
        self.updateAlgorithmParameters(Parameters={"nextStep":bool(nextStep)})
        if not nextStep: Operator.CM.clearCache()
        try:
            if   Executor == "YACS": self.__executeYACSScheme( SaveCaseInFile )
            else:                    self.__executePythonScheme( SaveCaseInFile )
        except Exception as e:
            if isinstance(e, SyntaxError): msg = "at %s: %s"%(e.offset, e.text)
            else: msg = ""
            raise ValueError(("during execution, the following error occurs:\n"+\
                             "\n%s %s\n\nSee also the potential messages, "+\
                             "which can show the origin of the above error, "+\
                             "in the launching terminal.\n")%(str(e),msg))
        return 0

    def __executePythonScheme(self, FileName=None):
        "Lancement du calcul"
        self.__case.register("executePythonScheme", dir(), locals())
        if FileName is not None:
            self.dump( FileName, "TUI")
        self.__adaoObject["AlgorithmParameters"].executePythonScheme( self.__adaoObject )
        return 0

    def __executeYACSScheme(self, FileName=None):
        "Lancement du calcul"
        self.__case.register("executeYACSScheme", dir(), locals())
        self.dump( FileName, "YACS")
        self.__adaoObject["AlgorithmParameters"].executeYACSScheme( FileName )
        return 0

    # -----------------------------------------------------------

    def dump(self, FileName=None, Formater="TUI"):
        "Restitution normalisée des commandes"
        __Upa = "\n".join(self.__PostAnalysis)
        return self.__case.dump(FileName, Formater, __Upa)

    def load(self, FileName=None, Content=None, Object=None, Formater="TUI"):
        "Chargement normalisé des commandes"
        __commands = self.__case.load(FileName, Content, Object, Formater)
        from numpy import array, matrix
        for __command in __commands:
            if __command.find("set")>-1 and __command.find("set_")<0:
                exec("self."+__command)
            else:
                self.__PostAnalysis.append(__command)
        return self

    def convert(self,
        FileNameFrom=None, ContentFrom=None, ObjectFrom=None, FormaterFrom="TUI",
        FileNameTo=None, FormaterTo="TUI",
        ):
        "Conversion normalisée des commandes"
        return self.load(
            FileName=FileNameFrom, Content=ContentFrom, Object=ObjectFrom, Formater=FormaterFrom
            ).dump(
            FileName=FileNameTo, Formater=FormaterTo
            )

    def clear(self):
        "Effacement du contenu du cas en cours"
        self.__init__(self.__name)

    # -----------------------------------------------------------

    def __with_directory(self, __filename=None):
        if os.path.exists(str(__filename)):
            __fullpath = __filename
        elif os.path.exists(os.path.join(str(self.__directory), str(__filename))):
            __fullpath = os.path.join(self.__directory, str(__filename))
        else:
            __fullpath = __filename
        return __fullpath

    def __dir__(self):
        "Clarifie la visibilité des méthodes"
        return ['set', 'get', 'execute', 'dump', 'load', '__doc__', '__init__', '__module__']

    def prepare_to_pickle(self):
        "Retire les variables non pickelisables, avec recopie efficace"
        if self.__adaoObject['AlgorithmParameters'] is not None:
            for k in self.__adaoObject['AlgorithmParameters'].keys():
                if k == "Algorithm": continue
                if k in self.__StoredInputs:
                    raise ValueError("The key \"%s\" to be transfered for pickling will overwrite an existing one."%(k,))
                if self.__adaoObject['AlgorithmParameters'].hasObserver( k ):
                    self.__adaoObject['AlgorithmParameters'].removeObserver( k, "", True )
                self.__StoredInputs[k] = self.__adaoObject['AlgorithmParameters'].pop(k, None)
        if sys.version_info[0] == 2:
            del self.__adaoObject # Because it breaks pickle in Python 2. Not required for Python 3
            del self.__case       # Because it breaks pickle in Python 2. Not required for Python 3
        if sys.version_info.major < 3:
            return 0
        else:
            return self.__StoredInputs

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#

from salome.kernel.logger import Logger
from salome.kernel import termcolor

adao_logger = Logger("ADAO")
adao_engine_logger = Logger("ADAO ENGINE")

def info(msg, logger = "ADAO"):

  if logger == "ADAO":
    adao_logger.setColor(termcolor.BLUE)
    adao_logger.info(msg)
  elif logger == "ENGINE":
    adao_engine_logger.setColor(termcolor.BLUE)
    adao_engine_logger.info(msg)

def debug(msg, logger = "ADAO"):

  if logger == "ADAO":
    adao_logger.setColor(termcolor.GREEN)
    adao_logger.debug(msg)
  elif logger == "ENGINE":
    adao_engine_logger.setColor(termcolor.GREEN)
    adao_engine_logger.debug(msg)

def error(msg, logger = "ADAO"):

  if logger == "ADAO":
    adao_logger.setColor(termcolor.RED)
    adao_logger.error(msg)
  elif logger == "ENGINE":
    adao_engine_logger.setColor(termcolor.RED)
    adao_engine_logger.error(msg)

def warning(msg, logger = "ADAO"):

  if logger == "ADAO":
    adao_logger.setColor(termcolor.CYAN)
    adao_logger.warning(msg)
  elif logger == "ENGINE":
    adao_engine_logger.setColor(termcolor.CYAN)
    adao_engine_logger.warning(msg)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import logging
from daCore import BasicObjects
import numpy

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "BLUE")
        self.defineRequiredParameter(
            name     = "StoreInternalVariables",
            default  = False,
            typecast = bool,
            message  = "Stockage des variables internes ou intermédiaires du calcul",
            )
        self.defineRequiredParameter(
            name     = "StoreSupplementaryCalculations",
            default  = [],
            typecast = tuple,
            message  = "Liste de calculs supplémentaires à stocker et/ou effectuer",
            listval  = [
                "Analysis",
                "APosterioriCorrelations",
                "APosterioriCovariance",
                "APosterioriStandardDeviations",
                "APosterioriVariances",
                "BMA",
                "CostFunctionJ",
                "CostFunctionJAtCurrentOptimum",
                "CostFunctionJb",
                "CostFunctionJbAtCurrentOptimum",
                "CostFunctionJo",
                "CostFunctionJoAtCurrentOptimum",
                "CurrentOptimum",
                "CurrentState",
                "Innovation",
                "MahalanobisConsistency",
                "OMA",
                "OMB",
                "SigmaBck2",
                "SigmaObs2",
                "SimulatedObservationAtBackground",
                "SimulatedObservationAtCurrentOptimum",
                "SimulatedObservationAtCurrentState",
                "SimulatedObservationAtOptimum",
                "SimulationQuantiles",
                ]
            )
        self.defineRequiredParameter(
            name     = "Quantiles",
            default  = [],
            typecast = tuple,
            message  = "Liste des valeurs de quantiles",
            minval   = 0.,
            maxval   = 1.,
            )
        self.defineRequiredParameter(
            name     = "SetSeed",
            typecast = numpy.random.seed,
            message  = "Graine fixée pour le générateur aléatoire",
            )
        self.defineRequiredParameter(
            name     = "NumberOfSamplesForQuantiles",
            default  = 100,
            typecast = int,
            message  = "Nombre d'échantillons simulés pour le calcul des quantiles",
            minval   = 1,
            )
        self.defineRequiredParameter(
            name     = "SimulationForQuantiles",
            default  = "Linear",
            typecast = str,
            message  = "Type de simulation pour l'estimation des quantiles",
            listval  = ["Linear", "NonLinear"]
            )
        self.requireInputArguments(
            mandatory= ("Xb", "Y", "HO", "R", "B"),
            )
        self.setAttributes(tags=(
            "DataAssimilation",
            "Linear",
            "Filter",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, R, B, Q)
        #
        Hm = HO["Tangent"].asMatrix(Xb)
        Hm = Hm.reshape(Y.size,Xb.size) # ADAO & check shape
        Ha = HO["Adjoint"].asMatrix(Xb)
        Ha = Ha.reshape(Xb.size,Y.size) # ADAO & check shape
        #
        # Utilisation éventuelle d'un vecteur H(Xb) précalculé
        # ----------------------------------------------------
        if HO["AppliedInX"] is not None and "HXb" in HO["AppliedInX"]:
            HXb = HO["AppliedInX"]["HXb"]
        else:
            HXb = Hm * Xb
        HXb = numpy.asmatrix(numpy.ravel( HXb )).T
        if Y.size != HXb.size:
            raise ValueError("The size %i of observations Y and %i of observed calculation H(X) are different, they have to be identical."%(Y.size,HXb.size))
        if max(Y.shape) != max(HXb.shape):
            raise ValueError("The shapes %s of observations Y and %s of observed calculation H(X) are different, they have to be identical."%(Y.shape,HXb.shape))
        #
        # Précalcul des inversions de B et R
        # ----------------------------------
        BI = B.getI()
        RI = R.getI()
        #
        # Calcul de l'innovation
        # ----------------------
        d  = Y - HXb
        #
        # Calcul de la matrice de gain et de l'analyse
        # --------------------------------------------
        if Y.size <= Xb.size:
            _A = R + numpy.dot(Hm, B * Ha)
            _u = numpy.linalg.solve( _A , d )
            Xa = Xb + B * Ha * _u
        else:
            _A = BI + numpy.dot(Ha, RI * Hm)
            _u = numpy.linalg.solve( _A , numpy.dot(Ha, RI * d) )
            Xa = Xb + _u
        self.StoredVariables["Analysis"].store( Xa.A1 )
        #
        # Calcul de la fonction coût
        # --------------------------
        if self._parameters["StoreInternalVariables"] or \
            self._toStore("CostFunctionJ")  or self._toStore("CostFunctionJAtCurrentOptimum") or \
            self._toStore("CostFunctionJb") or self._toStore("CostFunctionJbAtCurrentOptimum") or \
            self._toStore("CostFunctionJo") or self._toStore("CostFunctionJoAtCurrentOptimum") or \
            self._toStore("OMA") or \
            self._toStore("SigmaObs2") or \
            self._toStore("MahalanobisConsistency") or \
            self._toStore("SimulatedObservationAtCurrentOptimum") or \
            self._toStore("SimulatedObservationAtCurrentState") or \
            self._toStore("SimulatedObservationAtOptimum") or \
            self._toStore("SimulationQuantiles"):
            HXa = Hm * Xa
            oma = Y - HXa
        if self._parameters["StoreInternalVariables"] or \
            self._toStore("CostFunctionJ")  or self._toStore("CostFunctionJAtCurrentOptimum") or \
            self._toStore("CostFunctionJb") or self._toStore("CostFunctionJbAtCurrentOptimum") or \
            self._toStore("CostFunctionJo") or self._toStore("CostFunctionJoAtCurrentOptimum") or \
            self._toStore("MahalanobisConsistency"):
            Jb  = float( 0.5 * (Xa - Xb).T * BI * (Xa - Xb) )
            Jo  = float( 0.5 * oma.T * RI * oma )
            J   = Jb + Jo
            self.StoredVariables["CostFunctionJb"].store( Jb )
            self.StoredVariables["CostFunctionJo"].store( Jo )
            self.StoredVariables["CostFunctionJ" ].store( J )
            self.StoredVariables["CostFunctionJbAtCurrentOptimum"].store( Jb )
            self.StoredVariables["CostFunctionJoAtCurrentOptimum"].store( Jo )
            self.StoredVariables["CostFunctionJAtCurrentOptimum" ].store( J )
        #
        # Calcul de la covariance d'analyse
        # ---------------------------------
        if self._toStore("APosterioriCovariance") or \
            self._toStore("SimulationQuantiles"):
            if   (Y.size <= Xb.size): K  = B * Ha * (R + numpy.dot(Hm, B * Ha)).I
            elif (Y.size >  Xb.size): K = (BI + numpy.dot(Ha, RI * Hm)).I * Ha * RI
            A = B - K * Hm * B
            if min(A.shape) != max(A.shape):
                raise ValueError("The %s a posteriori covariance matrix A is of shape %s, despites it has to be a squared matrix. There is an error in the observation operator, please check it."%(self._name,str(A.shape)))
            if (numpy.diag(A) < 0).any():
                raise ValueError("The %s a posteriori covariance matrix A has at least one negative value on its diagonal. There is an error in the observation operator, please check it."%(self._name,))
            if logging.getLogger().level < logging.WARNING: # La verification n'a lieu qu'en debug
                try:
                    L = numpy.linalg.cholesky( A )
                except:
                    raise ValueError("The %s a posteriori covariance matrix A is not symmetric positive-definite. Please check your a priori covariances and your observation operator."%(self._name,))
            self.StoredVariables["APosterioriCovariance"].store( A )
        #
        # Calculs et/ou stockages supplémentaires
        # ---------------------------------------
        if self._parameters["StoreInternalVariables"] or self._toStore("CurrentState"):
            self.StoredVariables["CurrentState"].store( numpy.ravel(Xa) )
        if self._toStore("CurrentOptimum"):
            self.StoredVariables["CurrentOptimum"].store( numpy.ravel(Xa) )
        if self._toStore("Innovation"):
            self.StoredVariables["Innovation"].store( numpy.ravel(d) )
        if self._toStore("BMA"):
            self.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(Xa) )
        if self._toStore("OMA"):
            self.StoredVariables["OMA"].store( numpy.ravel(oma) )
        if self._toStore("OMB"):
            self.StoredVariables["OMB"].store( numpy.ravel(d) )
        if self._toStore("SigmaObs2"):
            TraceR = R.trace(Y.size)
            self.StoredVariables["SigmaObs2"].store( float( (d.T * (numpy.asmatrix(numpy.ravel(oma)).T)) ) / TraceR )
        if self._toStore("SigmaBck2"):
            self.StoredVariables["SigmaBck2"].store( float( (d.T * Hm * (Xa - Xb))/(Hm * B * Hm.T).trace() ) )
        if self._toStore("MahalanobisConsistency"):
            self.StoredVariables["MahalanobisConsistency"].store( float( 2.*J/d.size ) )
        if self._toStore("SimulationQuantiles"):
            nech = self._parameters["NumberOfSamplesForQuantiles"]
            YfQ  = None
            for i in range(nech):
                if self._parameters["SimulationForQuantiles"] == "Linear":
                    dXr = numpy.matrix(numpy.random.multivariate_normal(Xa.A1,A) - Xa.A1).T
                    dYr = numpy.matrix(numpy.ravel( Hm * dXr )).T
                    Yr = HXa + dYr
                elif self._parameters["SimulationForQuantiles"] == "NonLinear":
                    Xr = numpy.matrix(numpy.random.multivariate_normal(Xa.A1,A)).T
                    Yr = numpy.matrix(numpy.ravel( Hm * Xr )).T
                if YfQ is None:
                    YfQ = Yr
                else:
                    YfQ = numpy.hstack((YfQ,Yr))
            YfQ.sort(axis=-1)
            YQ = None
            for quantile in self._parameters["Quantiles"]:
                if not (0. <= float(quantile) <= 1.): continue
                indice = int(nech * float(quantile) - 1./nech)
                if YQ is None: YQ = YfQ[:,indice]
                else:          YQ = numpy.hstack((YQ,YfQ[:,indice]))
            self.StoredVariables["SimulationQuantiles"].store( YQ )
        if self._toStore("SimulatedObservationAtBackground"):
            self.StoredVariables["SimulatedObservationAtBackground"].store( numpy.ravel(HXb) )
        if self._toStore("SimulatedObservationAtCurrentState"):
            self.StoredVariables["SimulatedObservationAtCurrentState"].store( numpy.ravel(HXa) )
        if self._toStore("SimulatedObservationAtCurrentOptimum"):
            self.StoredVariables["SimulatedObservationAtCurrentOptimum"].store( numpy.ravel(HXa) )
        if self._toStore("SimulatedObservationAtOptimum"):
            self.StoredVariables["SimulatedObservationAtOptimum"].store( numpy.ravel(HXa) )
        #
        self._post_run(HO)
        return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



#-*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: André Ribes, andre.ribes@edf.fr, EDF R&D



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2020 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#

__author__="aribes/gboulant"

from PyQt5.QtGui import *
from PyQt5.QtCore import *
from PyQt5.QtWidgets import QApplication

import SalomePyQt
sgPyQt = SalomePyQt.SalomePyQt()

################################################
# GUI context class
# Used to store actions, menus, toolbars, etc...
################################################
# A gui context instance is created for each study. The dictionnary __study2context__
# keeps the mapping in memory. This context contains graphical objects that have
# to be created for each study. It contains at least the ui component builder that
# creates the menu and toolbar items (must be created for every study)

from daGuiImpl import adaoGuiHelper
from daGuiImpl.adaoGuiManager import AdaoCaseManager

class GUIcontext:
    adaoCaseManager = None
    def __init__(self):
        self.adaoCaseManager = AdaoCaseManager()

#~ __study2context__   = {}
#~ __current_context__ = None
__current_context__ = GUIcontext()
#~ def _setContext( studyID ):
    #~ global __study2context__, __current_context__
    #~ QApplication.processEvents()
    #~ if studyID not in __study2context__:
        #~ __study2context__[studyID] = GUIcontext()
        #~ pass
    #~ __current_context__ = __study2context__[studyID]
    #~ return __current_context__

# This object does not need to be embedded in a GUI context object. A single
# instance for all studies is a priori sufficient.

################################################
# Implementation of SALOME GUI interface
################################################

# called when module is initialized
# perform initialization actions
def initialize():
    pass

# called when module is initialized
# return map of popup windows to be used by the module
def windows():
    wm = {}
    wm[SalomePyQt.WT_ObjectBrowser] = Qt.LeftDockWidgetArea
    wm[SalomePyQt.WT_PyConsole]     = Qt.BottomDockWidgetArea
    return wm

# called when module is initialized
# return list of 2d/3d views to be used ny the module
def views():
  return []

def createPreferences():
    """
    Called when module is initialized. Export module's preferences.
    """
    pass

# called when module is activated
# returns True if activating is successfull and False otherwise
def activate():
    #~ ctx = _setContext( sgPyQt.getStudyId() )
    global __current_context__ ; ctx = __current_context__
    QApplication.processEvents()
    ctx.adaoCaseManager.activate()
    return True

# called when module is deactivated
def deactivate():
    #~ ctx = _setContext( sgPyQt.getStudyId() )
    global __current_context__ ; ctx = __current_context__
    QApplication.processEvents()
    ctx.adaoCaseManager.deactivate()

# called when active study is changed
# active study ID is passed as parameter
def activeStudyChanged(): # studyID ):
    #~ ctx = _setContext( sgPyQt.getStudyId() )
    QApplication.processEvents()

# called when popup menu is invoked
# popup menu and menu context are passed as parameters
def createPopupMenu( popup, context ):
  #~ activeStudyId = sgPyQt.getStudyId()
  #~ ctx = _setContext(sgPyQt.getStudyId())
  global __current_context__ ; ctx = __current_context__
  QApplication.processEvents()
  selcount, selected = adaoGuiHelper.getAllSelected()# activeStudyId)
  if selcount == 1:
    selectedItem = adaoGuiHelper.getSelectedItem()# activeStudyId)
    popup = ctx.adaoCaseManager.salome_manager.createPopupMenuOnItem(popup, selectedItem) # activeStudyId, selectedItem)

def OnGUIEvent(actionId) :
    """
    Called when an event is raised from a graphic item (click on menu item or
    toolbar button). The actionId value is the ID associated to the item.
    """
    #~ ctx = _setContext( sgPyQt.getStudyId() )
    global __current_context__ ; ctx = __current_context__
    QApplication.processEvents()
    ctx.adaoCaseManager.processGUIEvent(actionId)

# called when module's preferences are changed
# preference's resources section and setting name are passed as parameters
def preferenceChanged( section, setting ):
    pass

# called when active view is changed
# view ID is passed as parameter
def activeViewChanged( viewID ):
    pass

# called when active view is cloned
# cloned view ID is passed as parameter
def viewCloned( viewID ):
    pass

# called when active view is viewClosed
# view ID is passed as parameter
def viewClosed( viewID ):
    pass

